"""
Tests for L{xmantissa.websharing} and L{xmantissa.publicweb}.
"""
from zope.interface import Interface, Attribute, implements

from twisted.python.components import registerAdapter

from twisted.trial.unittest import TestCase

from nevow import rend, url
from nevow.athena import LiveElement

from epsilon.structlike import record

from axiom.item import Item
from axiom.attributes import integer, text
from axiom.store import Store
from axiom.userbase import LoginSystem, LoginMethod
from axiom.dependency import installOn
from axiom.plugins.mantissacmd import Mantissa

from xmantissa import (
    websharing, sharing, signup, offering, product, ixmantissa)



class _TemplateNameResolver(Item):
    """
    An L{ixmantissa.ITemplateNameResolver} with an implementation of
    L{getDocFactory} which doesn't require the presence any disk templates.
    """
    powerupInterfaces = (ixmantissa.ITemplateNameResolver,)

    magicTemplateName = text(doc="""
    L{magicDocFactoryValue} will be returned by L{getDocFactory} if it is passed
    this string as the first argument.""")

    magicDocFactoryValue = text(doc="""
    The string value to be returned from L{getDocFactory} when the name it is
    passed matches L{magicTemplateName}."""
    # if anything starts to care too much about what the docFactory is, we
    # won't be able to get away with just using a string.
    )

    # ITemplateNameResolver
    def getDocFactory(self, name, default=None):
        """
        If C{name} matches L{self.magicTemplateName}, return
        L{self.magicTemplateName}, otherwise return C{default}.
        """
        if name == self.magicTemplateName:
            return self.magicDocFactoryValue
        return default

class ITest(Interface):
    """
    Interface for L{TestAppPowerup} to be shared on.
    """
    store = Attribute("expose 'store' for testing")

class TestAppPowerup(Item):
    implements(ITest)
    attr = integer()

    def installed(self):
        """
        Share this item once installed.
        """
        shareid = u'test'
        sharing.getEveryoneRole(self.store
                        ).shareItem(self,
                          shareID=shareid)
        websharing.addDefaultShareID(self.store, shareid, 0)



class WebSharingTestCase(TestCase):
    """
    Tests for L{xmantissa.websharing.linkTo}
    """
    def setUp(self):
        """
        Set up some state.
        """
        self.s = Store()
        self.ls = LoginSystem(store=self.s)
        installOn(self.ls, self.s)
        acct = self.ls.addAccount(
            u'right', u'host', u'', verified=True, internal=True)
        acct.addLoginMethod(
            u'wrong', u'host', internal=False, verified=False)
        self.share = sharing.shareItem(self.ls, shareID=u'loginsystem')


    def test_noLoginMethods(self):
        """
        L{websharing.linkTo} raises a L{RuntimeError} when the shared item is
        in a store with no internal L{LoginMethod}s.
        """
        for lm in self.s.query(LoginMethod):
            lm.internal = False
        self.assertRaises(RuntimeError, websharing.linkTo, self.share)


    def test_linkToShare(self):
        """
        Test that L{xmantissa.websharing.linkTo} generates a URL using the
        localpart of the account's internal L{axiom.userbase.LoginMethod}
        """
        self._verifyPath(websharing.linkTo(self.share))


    def _verifyPath(self, linkURL):
        """
        Verify that the given url matches the test's expectations.
        """
        self.failUnless(isinstance(linkURL, url.URL),
                        "linkTo should return a nevow.url.URL, not %r" %
                        (type(linkURL)))
        self.assertEquals(str(linkURL), '/users/right/loginsystem')


    def test_linkToProxy(self):
        """
        Test that L{xmantissa.websharing.linkTo} generates a URL that I can
        link to.
        """
        self._verifyPath(
            websharing.linkTo(sharing.getShare(self.s, sharing.getEveryoneRole(
                        self.s), u'loginsystem')))


    def test_shareURLInjectsShareID(self):
        """
        Test that L{xmantissa.websharing._ShareURL} injects the share ID the
        constructor is passed when C{child} is called.
        """
        for (shareID, urlID) in [(u'a', 'a'), (u'\xe9', '%C3%A9')]:
            shareURL = websharing._ShareURL(shareID, netloc='', scheme='')
            self.assertEqual(str(shareURL.child('c')), '/%s/c' % urlID)
            # make sure subsequent child calls on the original have the same
            # behaviour
            self.assertEqual(str(shareURL.child('d')), '/%s/d' % urlID)
            # and that child calls on the returned urls don't (i.e. not
            # '/a/c/a/d'
            self.assertEqual(str(shareURL.child('c').child('d')),
                             '/%s/c/d' % urlID)


    def test_shareURLNoStoreID(self):
        """
        Test that L{xmantissa.websharing._ShareURL} behaves like a regular
        L{nevow.url.URL} when no store ID is passed.
        """
        shareURL = websharing._ShareURL(None, netloc='', scheme='')
        self.assertEqual(str(shareURL.child('a')), '/a')
        self.assertEqual(str(shareURL.child('a').child('b')), '/a/b')


    def test_shareURLNoClassmethodConstructors(self):
        """
        Verify that the C{fromRequest}, C{fromContext} and C{fromString}
        constructors on L{xmantissa.websharing._ShareURL} throw
        L{NotImplementedError}.
        """
        for meth in (websharing._ShareURL.fromRequest,
                     websharing._ShareURL.fromString,
                     websharing._ShareURL.fromContext):
            self.assertRaises(
                NotImplementedError,
                lambda: meth(None))


    def test_shareURLCloneMaintainsShareID(self):
        """
        Test that L{xmantissa.websharing._ShareURL} can be cloned, and that
        clones will remember the share ID.
        """
        shareURL = websharing._ShareURL(u'a', netloc='', scheme='')
        shareURL = shareURL.cloneURL('', '', None, None, '')
        self.assertEqual(shareURL._shareID, u'a')


    def test_defaultShareIDInteractionMatching(self):
        """
        Verify that L{websharing.linkTo} does not explicitly include a share
        ID in the URL if the ID of the share it is passed matches the default.
        """
        websharing.addDefaultShareID(self.s, u'share-id', 0)
        sharing.shareItem(Shareable(store=self.s), shareID=u'share-id')
        share = sharing.getShare(
            self.s, sharing.getEveryoneRole(self.s), u'share-id')
        url = websharing.linkTo(share)
        self.assertEqual(str(url), '/users/right/')
        # and if we call child()
        self.assertEqual(str(url.child('child')), '/users/right/share-id/child')


    def test_defaultShareIDInteractionNoMatch(self):
        """
        Verify that L{websharing.linkTo} explicitly includes a share ID in the
        URL if the ID of the share it is passed doesn't match the default.
        """
        websharing.addDefaultShareID(self.s, u'share-id', 0)
        shareable = Shareable(store=self.s)
        sharing.shareItem(Shareable(store=self.s), shareID=u'not-the-share-id')
        share = sharing.getShare(
            self.s, sharing.getEveryoneRole(self.s), u'not-the-share-id')
        url = websharing.linkTo(share)
        self.assertEqual(str(url), '/users/right/not-the-share-id')


    def test_appStoreLinkTo(self):
        """
        When L{websharing.linkTo} is called on a shared item in an app store,
        it returns an URL with a single path segment consisting of the app's
        name.
        """
        s = Store(dbdir=self.mktemp())
        Mantissa().installSite(s, u"localhost", u"", False)
        Mantissa().installAdmin(s, u'admin', u'localhost', u'asdf')
        off = offering.Offering(
            name=u'test_offering',
            description=u'Offering for creating a sample app store',
            siteRequirements=[],
            appPowerups=[TestAppPowerup],
            installablePowerups=[],
            loginInterfaces=[],
            themes=[],
            )
        userbase = s.findUnique(LoginSystem)
        adminAccount = userbase.accountByAddress(u'admin', u'localhost')
        conf = adminAccount.avatars.open().findUnique(
            offering.OfferingConfiguration)
        conf.installOffering(off, None)
        ss = userbase.accountByAddress(off.name, None).avatars.open()
        sharedItem = sharing.getEveryoneRole(ss).getShare(
            websharing.getDefaultShareID(ss))
        linkURL = websharing.linkTo(sharedItem)
        self.failUnless(isinstance(linkURL, url.URL),
                        "linkTo should return a nevow.url.URL, not %r" %
                        (type(linkURL)))
        self.assertEquals(str(linkURL), '/test_offering/')


class _UserIdentificationMixin:
    def setUp(self):
        self.siteStore = Store(filesdir=self.mktemp())
        Mantissa().installSite(self.siteStore, u"localhost", u"", False)
        Mantissa().installAdmin(self.siteStore, u'admin', u'localhost', '')
        self.loginSystem = self.siteStore.findUnique(LoginSystem)
        self.adminStore = self.loginSystem.accountByAddress(
            u'admin', u'localhost').avatars.open()
        sc = self.adminStore.findUnique(signup.SignupConfiguration)
        self.signup = sc.createSignup(
            u'testuser@localhost',
            signup.UserInfoSignup,
            {'prefixURL': u''},
            product.Product(store=self.siteStore, types=[]),
            u'', u'')



class UserIdentificationTestCase(_UserIdentificationMixin, TestCase):
    """
    Tests for L{xmantissa.websharing._storeFromUsername}
    """
    def test_sameLocalpartAndUsername(self):
        """
        Test that L{xmantissa.websharing._storeFromUsername} doesn't get
        confused when the username it is passed is the same as the localpart
        of that user's email address
        """
        self.signup.createUser(
            u'', u'username', u'localhost', u'', u'username@internet')
        self.assertIdentical(
            websharing._storeFromUsername(self.siteStore, u'username'),
            self.loginSystem.accountByAddress(
                u'username', u'localhost').avatars.open())


    def test_usernameMatchesOtherLocalpart(self):
        """
        Test that L{xmantissa.websharing._storeFromUsername} doesn't get
        confused when the username it is passed matches the localpart of
        another user's email address
        """
        self.signup.createUser(
            u'', u'username', u'localhost', u'', u'notusername@internet')
        self.signup.createUser(
            u'', u'notusername', u'localhost', u'', u'username@internet')
        self.assertIdentical(
            websharing._storeFromUsername(self.siteStore, u'username'),
            self.loginSystem.accountByAddress(
                u'username', u'localhost').avatars.open())


class IShareable(Interface):
    """
    Dummy interface for Shareable.
    """
    magicValue = Attribute(
        """
        A magical value.
        """)


    fragmentName = Attribute(
        """
        The value that the corresponding L{ShareableView} should use for its
        C{fragmentName} attribute.
        """)



class Shareable(Item):
    """
    This is a dummy class that may be shared.
    """
    implements(IShareable)

    magicValue = integer()



class ShareableView(LiveElement):
    """
    Nothing to see here, move along.

    @ivar customizedFor: The username we were customized for, or C{None}.
    """
    implements(ixmantissa.INavigableFragment,
               ixmantissa.ICustomizable)

    customizedFor = None
    fragmentName = 'bogus'

    def __init__(self, shareable):
        """
        adapt a shareable to INavigableFragment
        """
        super(ShareableView, self).__init__()
        self.shareable = shareable


    def showMagicValue(self):
        """
        retrieve the magic value from my model
        """
        return self.shareable.magicValue


    # XXX: Everything below in this class should not be required.  It's here to
    # satisfy implicit requirements in SharingIndex.locateChild, but there
    # should be test coverage ensuring that it is not required and that
    # customizeFor is only invoked if you provide the ICustomizable interface.

    def customizeFor(self, user):
        """
        Customize me by returning myself, and storing the username we were
        customized for as L{self.customizedFor}.
        """
        self.customizedFor = user
        return self



registerAdapter(ShareableView, IShareable,
                ixmantissa.INavigableFragment)


class FakePage(record('wrappedFragment')):
    """
    A fake page that simply contains a wrapped fragment; analagous to the
    various shell page classes in the tests which use it.
    """



class FakeShellFactory(record('username')):
    """
    An L{IWebViewer} which returns a fake shell page and maps a role
    directly to its given username.
    """

    def wrapModel(self, model):
        """
        Adapt the given model to L{INavigableFragment} and then wrap it in a
        L{FakePage}.
        """
        return FakePage(ixmantissa.INavigableFragment(model))


    def roleIn(self, userStore):
        """
        Return the primary role for the username passed to me.
        """
        return sharing.getPrimaryRole(userStore, self.username)



class UserIndexPageTestCase(_UserIdentificationMixin, TestCase):
    """
    Tests for L{xmantissa.websharing.UserIndexPage}
    """
    username = u'alice'
    domain = u'example.com'
    shareID = u'ashare'

    def setUp(self):
        """
        Create an additional user for UserIndexPage, and share a single item with a
        shareID of the empty string.
        """
        _UserIdentificationMixin.setUp(self)
        self.magicValue = 123412341234
        self.signup.createUser(
            u'', self.username, self.domain, u'', u'username@internet')
        self.userStore = websharing._storeFromUsername(
            self.siteStore, self.username)
        self.shareable = Shareable(store=self.userStore,
                                   magicValue=self.magicValue)
        self.share = sharing.shareItem(self.shareable, shareID=self.shareID)


    def makeSharingIndex(self, username):
        """
        Create a L{SharingIndex}.
        """
        return websharing.SharingIndex(
            self.userStore,
            webViewer=FakeShellFactory(username))


    def test_locateChild(self):
        """
        L{websharing.UserIndexPage.locateChild} should return the named user's
        L{websharing.SharingIndex} (and any remaining segments), or
        L{rend.NotFound}.
        """
        # Test against at least one other valid user.
        self.signup.createUser(
            u'Andr\xe9', u'andr\xe9', u'localhost', u'', u'andr\xe9@internet')
        userStore2 = websharing._storeFromUsername(self.siteStore, u'andr\xe9')
        index = websharing.UserIndexPage(self.loginSystem, FakeShellFactory(None))

        for _username, _store in [(self.username, self.userStore),
                                  (u'andr\xe9', userStore2)]:
            (found, remaining) = index.locateChild(
                None, [_username.encode('utf-8'), 'x', 'y', 'z'])

            self.assertTrue(isinstance(found, websharing.SharingIndex))
            self.assertIdentical(found.userStore, _store)
            self.assertEquals(remaining, ['x', 'y', 'z'])

        self.assertIdentical(index.locateChild(None, ['bogus', 'username']),
                             rend.NotFound)


    def test_linkToMatchesUserURL(self):
        """
        Test that L{xmantissa.websharing.linkTo} generates a URL using the
        localpart of the account's internal L{axiom.userbase.LoginMethod}
        """
        pathString = str(websharing.linkTo(self.share))
        expected = u'/users/%s/%s' % (self.username, self.shareID)
        self.assertEqual(pathString, expected.encode('ascii'))


    def test_emptySegmentNoDefault(self):
        """
        Verify that we get L{rend.NotFound} from
        L{websharing.SharingIndex.locateChild} if there is no default share ID
        and we access the empty child.
        """
        sharingIndex = self.makeSharingIndex(None)
        result = sharingIndex.locateChild(None, ('',))
        self.assertIdentical(result, rend.NotFound)


    def test_emptySegmentWithDefault(self):
        """
        Verify that we get the right resource and segments from
        L{websharing.SharingIndex.locateChild} if there is a default share ID
        and we access the empty child.
        """
        websharing.addDefaultShareID(self.userStore, u'ashare', 0)
        sharingIndex = self.makeSharingIndex(None)
        SEGMENTS = ('', 'foo', 'bar')
        (res, segments) = sharingIndex.locateChild(None, SEGMENTS)
        self.assertEqual(
            res.wrappedFragment.showMagicValue(), self.magicValue)
        self.assertEqual(segments, SEGMENTS[1:])


    def test_invalidShareIDNoDefault(self):
        """
        Verify that we get L{rend.NotFound} from
        L{websharing.SharingIndex.locateChild} if there is no default share ID
        and we access an invalid segment.
        """
        sharingIndex = self.makeSharingIndex(None)
        result = sharingIndex.locateChild(None, ('foo',))
        self.assertIdentical(result, rend.NotFound)


    def test_validShareID(self):
        """
        Verify that we get the right resource and segments from
        L{websharing.SharingIndex.locateChild} if we access a valid share ID.
        """
        websharing.addDefaultShareID(self.userStore, u'', 0)

        otherShareable = Shareable(store=self.userStore,
                                   magicValue=self.magicValue + 3)

        for _shareID in [u'foo', u'f\xf6\xf6']:
            otherShare = sharing.shareItem(otherShareable, shareID=_shareID)
            sharingIndex = self.makeSharingIndex(None)
            SEGMENTS = (_shareID.encode('utf-8'), 'bar')
            (res, segments) = sharingIndex.locateChild(None, SEGMENTS)
            self.assertEqual(
                res.wrappedFragment.showMagicValue(), self.magicValue + 3)
            self.assertEqual(segments, SEGMENTS[1:])



class DefaultShareIDTestCase(TestCase):
    """
    Tests for L{websharing.addDefaultShareID} and L{websharing.getDefaultShareID}.
    """
    def test_createsItem(self):
        """
        Verify that L{websharing.addDefaultShareID} creates a
        L{websharing._DefaultShareID} item.
        """
        store = Store()
        websharing.addDefaultShareID(store, u'share id', -22)
        item = store.findUnique(websharing._DefaultShareID)
        self.assertEqual(item.shareID, u'share id')
        self.assertEqual(item.priority, -22)


    def test_findShareID(self):
        """
        Verify that L{websharing.getDefaultShareID} reads the share ID set by a
        L{websharing.addDefaultShareID} call.
        """
        store = Store()
        websharing.addDefaultShareID(store, u'share id', 0)
        self.assertEqual(websharing.getDefaultShareID(store), u'share id')


    def test_findHighestPriorityShareID(self):
        """
        Verify that L{websharing.getDefaultShareID} reads the highest-priority
        share ID set by L{websharing.addDefaultShareID}.
        """
        store = Store()
        websharing.addDefaultShareID(store, u'share id!', 24)
        websharing.addDefaultShareID(store, u'share id',  25)
        websharing.addDefaultShareID(store, u'share id.', -1)
        self.assertEqual(websharing.getDefaultShareID(store), u'share id')


    def test_findsNoItem(self):
        """
        Verify that L{websharing.getDefaultShareID} returns C{u''} if there is
        no default share ID.
        """
        self.assertEqual(websharing.getDefaultShareID(Store()), u'')



# -*- test-case-name: xmantissa.test.test_signup,xmantissa.test.test_password_reset -*-

import os, rfc822, md5, time, random
from itertools import chain

from zope.interface import Interface, implements

from twisted.cred.portal import IRealm
from twisted.python.components import registerAdapter
from twisted.mail import smtp, relaymanager
from twisted.python.util import sibpath
from twisted.python import log
from twisted import plugin

from epsilon import extime

from axiom.item import Item, transacted, declareLegacyItem
from axiom.attributes import integer, reference, text, timestamp, AND
from axiom.iaxiom import IBeneficiary
from axiom import userbase, upgrade
from axiom.userbase import getDomainNames
from axiom.dependency import installOn

from nevow.rend import Page, NotFound
from nevow.url import URL
from nevow.inevow import IResource, ISession
from nevow import inevow, tags, athena
from nevow.athena import expose

from xmantissa.ixmantissa import (
    ISiteRootPlugin, IStaticShellContent, INavigableElement,
    INavigableFragment, ISignupMechanism, ITemplateNameResolver)
from xmantissa.website import PrefixURLMixin, WebSite
from xmantissa.websession import usernameFromRequest
from xmantissa.publicweb import PublicAthenaLivePage, PublicPage
from xmantissa.webnav import Tab
from xmantissa.webtheme import ThemedDocumentFactory, getLoader
from xmantissa.webapp import PrivateApplication
from xmantissa import plugins, liveform
from xmantissa.websession import PersistentSession
from xmantissa.smtp import parseAddress
from xmantissa.error import ArgumentError
from xmantissa.product import Product


_theMX = None
def getMX():
    """
    Retrieve the single MXCalculator instance, creating it first if
    necessary.
    """
    global _theMX
    if _theMX is None:
        _theMX = relaymanager.MXCalculator()
    return _theMX



def _sendEmail(_from, to, msg):

    def gotMX(mx):
        return smtp.sendmail(str(mx.name), _from, [to], msg)

    return getMX().getMX(to.split('@', 1)[1]).addCallback(gotMX)



class PasswordResetResource(PublicPage):
    """
    I handle the user-facing parts of password reset - the web form junk and
    sending of emails.

    The user sees a password reset form. The form asks the user for the
    username they use on this server. The form data is posted back to the
    C{PasswordResetResource}. On render, the resource creates a new 'attempt'
    (a L{_PasswordResetAttempt}) and sends an email to one of user's external
    addresses, if such a thing exists. The email contains a URL which is a
    child of this page, the last segment being the attempt ID.

    The user can then check their external email account and follow the link.
    Clicking the link loads the stored attempt and presents a password change
    form. The user can then specify a new password, click submit and their
    password will be reset.

    Currently unspecified behavior:
    - What happens when the provided username doesn't exist.
    - What happens when the provided passwords mismatch.
    - What happens when the user doesn't have an external account registered.

    @ivar store: a site store containing a L{WebSite}.
    @type store: L{axiom.store.Store}

    @ivar templateResolver: a template resolver instance that will return
        the appropriate doc factory.
    """

    attempt = None

    def __init__(self, store, templateResolver=None):
        if templateResolver is None:
            templateResolver = ITemplateNameResolver(store)
        PublicPage.__init__(self, None, store,
                            templateResolver.getDocFactory('reset'),
                            None, None, templateResolver)
        self.store = store
        self.loginSystem = store.findUnique(userbase.LoginSystem, default=None)


    def locateChild(self, ctx, segments):
        """
        Initialize self with the given key's L{_PasswordResetAttempt}, if any.

        @param segments: a L{_PasswordResetAttempt} key (hopefully)
        @return: C{(self, ())} with C{self.attempt} initialized, or L{NotFound}
        @see: L{attemptByKey}
        """
        if len(segments) == 1:
            attempt = self.attemptByKey(unicode(segments[0]))
            if attempt is not None:
                self.attempt = attempt
                return (self, ())
        return NotFound


    def renderHTTP(self, ctx):
        """
        Handle the password reset form.

        The following exchange describes the process:

            S: Render C{reset}
            C: POST C{username} or C{email}
            S: L{handleRequestForUser}, render C{reset-check-email}

            (User follows the emailed reset link)

            S: Render C{reset-step-two}
            C: POST C{password1}
            S: L{resetPassword}, render C{reset-done}
        """
        req = inevow.IRequest(ctx)

        if req.method == 'POST':
            if req.args.get('username', [''])[0]:
                user = unicode(usernameFromRequest(req), 'ascii')
                self.handleRequestForUser(user, URL.fromContext(ctx))
                self.fragment = self.templateResolver.getDocFactory(
                    'reset-check-email')
            elif req.args.get('email', [''])[0]:
                email = req.args['email'][0].decode('ascii')
                acct = self.accountByAddress(email)
                if acct is not None:
                    username = '@'.join(
                        userbase.getAccountNames(acct.avatars.open()).next())
                    self.handleRequestForUser(username, URL.fromContext(ctx))
                self.fragment = self.templateResolver.getDocFactory('reset-check-email')
            elif 'password1' in req.args:
                (password,) = req.args['password1']
                self.resetPassword(self.attempt, unicode(password))
                self.fragment = self.templateResolver.getDocFactory('reset-done')
            else:
                # Empty submit;  redirect back to self
                return URL.fromContext(ctx)
        elif self.attempt:
            self.fragment = self.templateResolver.getDocFactory('reset-step-two')

        return PublicPage.renderHTTP(self, ctx)


    def handleRequestForUser(self, username, url):
        """
        User C{username} wants to reset their password.  Create an attempt
        item, and send them an email if the username is valid
        """
        attempt = self.newAttemptForUser(username)
        account = self.accountByAddress(username)
        if account is None:
            # do we want to disclose this to the user?
            return
        email = self.getExternalEmail(account)
        if email is not None:
            self.sendEmail(url, attempt, email)


    def sendEmail(self, url, attempt, email, _sendEmail=_sendEmail):
        """
        Send an email for the given L{_PasswordResetAttempt}.

        @type url: L{URL}
        @param url: The URL of the password reset page.

        @type attempt: L{_PasswordResetAttempt}
        @param attempt: An L{Item} representing a particular user's attempt to
        reset their password.

        @type email: C{str}
        @param email: The email will be sent to this address.
        """

        host = url.netloc.split(':', 1)[0]
        from_ = 'reset@' + host

        body = file(sibpath(__file__, 'reset.rfc2822')).read()
        body %= {'from': from_,
                 'to': email,
                 'date': rfc822.formatdate(),
                 'message-id': smtp.messageid(),
                 'link': url.child(attempt.key)}

        _sendEmail(from_, email, body)


    def attemptByKey(self, key):
        """
        Locate the L{_PasswordResetAttempt} that corresponds to C{key}
        """

        return self.store.findUnique(_PasswordResetAttempt,
                                     _PasswordResetAttempt.key == key,
                                     default=None)


    def _makeKey(self, usern):
        """
        Make a new, probably unique key. This key will be sent in an email to
        the user and is used to access the password change form.
        """
        return unicode(md5.new(str((usern, time.time(), random.random()))).hexdigest())


    def newAttemptForUser(self, user):
        """
        Create an L{_PasswordResetAttempt} for the user whose username is C{user}
        @param user: C{unicode} username
        """
        # we could query for other attempts by the same
        # user within some timeframe and raise an exception,
        # if we wanted
        return _PasswordResetAttempt(store=self.store,
                                     username=user,
                                     timestamp=extime.Time(),
                                     key=self._makeKey(user))


    def getExternalEmail(self, account):
        """
        @return: str which is an external email address for the C{account}.
        None if there is no such address.
        """
        # XXX - shouldn't userbase do this for me? - jml
        method = self.store.findFirst(
            userbase.LoginMethod,
            AND (userbase.LoginMethod.account == account,
                 userbase.LoginMethod.internal == False))
        if method is None:
            return None
        else:
            return '%s@%s' % (method.localpart, method.domain)


    def accountByAddress(self, username):
        """
        @return: L{userbase.LoginAccount} for C{username} or None
        """
        userAndDomain = username.split('@', 1)
        if len(userAndDomain) != 2:
            return None
        return self.loginSystem.accountByAddress(*userAndDomain)


    def resetPassword(self, attempt, newPassword):
        """
        @param attempt: L{_PasswordResetAttempt}

        reset the password of the user who initiated C{attempt} to C{newPassword},
        and afterward, delete the attempt and any persistent sessions that belong
        to the user
        """

        self.accountByAddress(attempt.username).password = newPassword

        self.store.query(
            PersistentSession,
            PersistentSession.authenticatedAs == str(attempt.username)
            ).deleteFromStore()

        attempt.deleteFromStore()



class _PasswordResetAttempt(Item):
    """
    I represent as as-yet incomplete attempt at password reset
    """

    typeName = 'password_reset_attempt'
    schemaVersion = 1

    key = text()
    username = text()
    timestamp = timestamp()



class PasswordReset(Item):
    """
    I was an item that contained some of the model functionality of
    L{PasswordResetResource}, but now I am just a shell item (see
    L{passwordReset1to2}) and my functionality has been moved to
    L{PasswordResetResource}.
    """
    typeName = 'password_reset'
    schemaVersion = 2

    installedOn = reference()



def passwordReset1to2(old):
    """
    Power down and delete the item
    """
    new = old.upgradeVersion(old.typeName, 1, 2, installedOn=None)
    for iface in new.store.interfacesFor(new):
        new.store.powerDown(new, iface)
    new.deleteFromStore()

upgrade.registerUpgrader(passwordReset1to2, 'password_reset', 1, 2)



class NoSuchFactory(Exception):
    """
    An attempt was made to create a signup page using the name of a benefactor
    factory which did not correspond to anything in the database.
    """



class TicketClaimer(Page):
    def childFactory(self, ctx, name):
        for T in self.original.store.query(
            Ticket,
            AND(Ticket.booth == self.original,
                Ticket.nonce == unicode(name, 'ascii'))):
            something = T.claim()
            res = IResource(something)
            lgo = getattr(res, 'logout', lambda : None)
            ISession(ctx).setDefaultResource(res, lgo)
            return URL.fromContext(ctx).click("/private")
        return None



class TicketBooth(Item, PrefixURLMixin):
    implements(ISiteRootPlugin)

    typeName = 'ticket_powerup'
    schemaVersion = 1

    sessioned = True

    claimedTicketCount = integer(default=0)
    createdTicketCount = integer(default=0)

    defaultTicketEmail = text(default=None)

    prefixURL = 'ticket'

    def createResource(self):
        return TicketClaimer(self)

    def createTicket(self, issuer, email, product):
        t = self.store.findOrCreate(
            Ticket,
            product=product,
            booth=self,
            avatar=None,
            issuer=issuer,
            email=email)
        return t

    createTicket = transacted(createTicket)

    def ticketClaimed(self, ticket):
        self.claimedTicketCount += 1

    def ticketLink(self, domainName, httpPortNumber, nonce):
        httpPort = ''
        httpScheme = 'http'

        if httpPortNumber == 443:
            httpScheme = 'https'
        elif httpPortNumber != 80:
            httpPort = ':' + str(httpPortNumber)

        return '%s://%s%s/%s/%s' % (
            httpScheme, domainName, httpPort, self.prefixURL, nonce)

    def issueViaEmail(self, issuer, email, product, templateData,
                      domainName, httpPort=80):
        """
        Send a ticket via email to the supplied address, which, when claimed, will
        create an avatar and allow the given product to endow it with
        things.

        @param issuer: An object, preferably a user, to track who issued this
        ticket.

        @param email: a str, formatted as an rfc2821 email address
        (user@domain) -- source routes not allowed.

        @param product: an instance of L{Product}

        @param domainName: a domain name, used as the domain part of the
        sender's address, and as the web server to generate a link to within
        the email.

        @param httpPort: a port number for the web server running on domainName

        @param templateData: A string containing an rfc2822-format email
        message, which will have several python values interpolated into it
        dictwise:

            %(from)s: To be used for the From: header; will contain an
             rfc2822-format address.

            %(to)s: the address that we are going to send to.

            %(date)s: an rfc2822-format date.

            %(message-id)s: an rfc2822 message-id

            %(link)s: an HTTP URL that we are generating a link to.

        """

        ticket = self.createTicket(issuer,
                                   unicode(email, 'ascii'),
                                   product)
        nonce = ticket.nonce

        signupInfo = {'from': 'signup@'+domainName,
                      'to': email,
                      'date': rfc822.formatdate(),
                      'message-id': smtp.messageid(),
                      'link': self.ticketLink(domainName, httpPort, nonce)}

        msg = templateData % signupInfo

        return ticket, _sendEmail(signupInfo['from'], email, msg)



def _generateNonce():
    return unicode(os.urandom(16).encode('hex'), 'ascii')



class ITicketIssuer(Interface):
    def issueTicket(emailAddress):
        pass



class SignupMechanism(object):
    """
    I am a Twisted plugin helper.

    Instantiate me at module scope in a xmantissa.plugins submodule, including
    a name and description for the administrator.
    """
    implements(ISignupMechanism, plugin.IPlugin)
    def __init__(self, name, description, itemClass, configuration):
        """

        @param name: the name (a short string) to display to the administrator
        for selecting this signup mechanism.

        @param description: the description (a long string) for this signup
        mechanism.

        @param itemClass: a reference to a callable which takes keyword
        arguments described by L{configuration}, in addition to:

            store: the store to create the item in

            booth: a reference to a L{TicketBooth} that can create tickets for
            the created signup mechanism

            product: the product being installed by this signup.

            emailTemplate: a template for the email to be sent to the user

            prompt: a short unicode string describing this signup mechanism, as
            distinct from others.  For example: "Student Sign Up", or "Faculty
            Sign Up"

        @param configuration: a list of LiveForm arguments.
        """
        self.name = name
        self.description = description
        self.itemClass = itemClass
        self.configuration = configuration



freeTicketSignupConfiguration = [
    liveform.Parameter('prefixURL',
                       liveform.TEXT_INPUT,
                       unicode,
                       u'The web location at which users will be able to request tickets.',
                       default=u'signup')]

class FreeTicketSignup(Item, PrefixURLMixin):
    implements(ISiteRootPlugin)

    typeName = 'free_signup'
    schemaVersion = 6

    sessioned = True

    prefixURL = text(allowNone=False)
    booth = reference()
    product = reference(doc="An instance of L{product.Product} to install on"
                        " the new user's store")
    emailTemplate = text()
    prompt = text()

    def createResource(self):
        return PublicAthenaLivePage(
            self.store,
            getLoader("signup"),
            IStaticShellContent(self.store, None),
            None,
            iface = ITicketIssuer,
            rootObject = self)

    def issueTicket(self, url, emailAddress):
        domain, port = url.get('hostname'), int(url.get('port') or 80)
        if os.environ.get('CC_DEV'):
            ticket = self.booth.createTicket(self, emailAddress, self.product)
            return '<a href="%s">Claim Your Account</a>' % (
                    self.booth.ticketLink(domain, port, ticket.nonce),)
        else:
            ticket, issueDeferred = self.booth.issueViaEmail(
                self,
                emailAddress.encode('ascii'), # heh
                self.product,
                self.emailTemplate,
                domain,
                port)

            issueDeferred.addCallback(
                lambda result: u'Please check your email for a ticket!')

            return issueDeferred



def freeTicketSignup1To2(old):
    return old.upgradeVersion('free_signup', 1, 2,
                              prefixURL=old.prefixURL,
                              booth=old.booth,
                              benefactor=old.benefactor)

upgrade.registerUpgrader(freeTicketSignup1To2, 'free_signup', 1, 2)



def freeTicketSignup2To3(old):
    emailTemplate = file(sibpath(__file__, 'signup.rfc2822')).read()
    emailTemplate %= {'blurb': u'',
                      'subject': 'Welcome to a Generic Axiom Application!',
                      'linktext': "Click here to claim your 'generic axiom application' account"}

    return old.upgradeVersion('free_signup', 2, 3,
                              prefixURL=old.prefixURL,
                              booth=old.booth,
                              benefactor=old.benefactor,
                              emailTemplate=emailTemplate)

upgrade.registerUpgrader(freeTicketSignup2To3, 'free_signup', 2, 3)

declareLegacyItem(typeName='free_signup',
                  schemaVersion=3,
                  attributes=dict(prefixURL=text(),
                                  booth=reference(),
                                  benefactor=reference(),
                                  emailTemplate=text()))



def freeTicketSignup3To4(old):
    return old.upgradeVersion('free_signup', 3, 4,
                              prefixURL=old.prefixURL,
                              booth=old.booth,
                              benefactor=old.benefactor,
                              emailTemplate=old.emailTemplate,
                              prompt=u'Sign Up')

upgrade.registerUpgrader(freeTicketSignup3To4, 'free_signup', 3, 4)

declareLegacyItem(typeName='free_signup',
                  schemaVersion=4,
                  attributes=dict(prefixURL=text(),
                                  booth=reference(),
                                  benefactor=reference(),
                                  emailTemplate=text(),
                                  prompt=text()))

def freeTicketSignup4To5(old):
    return old.upgradeVersion('free_signup', 4, 5,
                              prefixURL=old.prefixURL,
                              booth=old.booth,
                              benefactor=old.benefactor,
                              emailTemplate=old.emailTemplate,
                              prompt=old.prompt)

upgrade.registerUpgrader(freeTicketSignup4To5, 'free_signup', 4, 5)


declareLegacyItem(typeName='free_signup',
                  schemaVersion=5,
                  attributes=dict(prefixURL=text(),
                                  booth=reference(),
                                  benefactor=reference(),
                                  emailTemplate=text(),
                                  prompt=text()))

def freeTicketSignup5To6(old):
    newProduct = old.store.findOrCreate(Product,
                                        types=list(
        chain(*[b.powerupNames for b in
                old.benefactor.benefactors('ascending')])))
    return old.upgradeVersion('free_signup', 5, 6,
                              prefixURL=old.prefixURL,
                              booth=old.booth,
                              product=newProduct,
                              emailTemplate=old.emailTemplate,
                              prompt=old.prompt)

upgrade.registerUpgrader(freeTicketSignup5To6, "free_signup", 5, 6)

class ValidatingSignupForm(liveform.LiveForm):
    jsClass = u'Mantissa.Validate.SignupForm'

    _parameterNames = [
        'realName',
        'username',
        'domain',
        'password',
        'emailAddress']

    docFactory = ThemedDocumentFactory("user-info-signup", "templateResolver")

    def __init__(self, uis):
        self.userInfoSignup = uis
        self.templateResolver = ITemplateNameResolver(uis.store)
        super(ValidatingSignupForm, self).__init__(
            uis.createUser,
            [liveform.Parameter(pname, liveform.TEXT_INPUT, unicode)
             for pname in
             self._parameterNames])


    def getInitialArguments(self):
        """
        Retrieve a domain name from the user info signup item and return it so
        the client will know what domain it can sign up in.
        """
        return (self.userInfoSignup.getAvailableDomains()[0],)


    def usernameAvailable(self, username, domain):
        return self.userInfoSignup.usernameAvailable(username, domain)
    athena.expose(usernameAvailable)



class UserInfo(Item):
    """
    An Item which stores information gleaned from the signup process of
    L{UserInfoSignup}.  The L{UserInfo} will reside in the substore of the
    user (which was created during the signup process), and will record
    information about its owner.
    """
    schemaVersion = 2
    realName = text(
        doc="""
        The name entered at signup time by the user as their I{real} name.
        """)


def upgradeUserInfo1to2(oldUserInfo):
    """
    Concatenate the I{firstName} and I{lastName} attributes from the old user
    info item and set the result as the I{realName} attribute of the upgraded
    item.
    """
    newUserInfo = oldUserInfo.upgradeVersion(
        UserInfo.typeName, 1, 2,
        realName=oldUserInfo.firstName + u" " + oldUserInfo.lastName)
    return newUserInfo
upgrade.registerUpgrader(upgradeUserInfo1to2, UserInfo.typeName, 1, 2)



class UserInfoSignup(Item, PrefixURLMixin):
    """
    This signup page provides a way to sign up while including some relevant
    information about yourself, including the selection of a username.
    """

    implements(ISiteRootPlugin)

    powerupInterfaces = (ISiteRootPlugin,)
    schemaVersion = 2
    sessioned = True

    booth = reference()
    product = reference(
        doc="""
        An instance of L{product.Product} to install on the new user's store.
        """)
    emailTemplate = text()
    prompt = text()

    # ISiteRootPlugin

    prefixURL = text(allowNone=False)

    def createResource(self):
        page = PublicAthenaLivePage(
            self.store,
            ValidatingSignupForm(self),
            IStaticShellContent(self.store, None))
        page.needsSecure = True
        return page


    # UserInfoSignup
    def getAvailableDomains(self):
        """
        Return a list of domain names available on this site.
        """
        return getDomainNames(self.store)


    def usernameAvailable(self, username, domain):
        """
        Check to see if a username is available for the user to select.
        """
        if len(username) < 2:
            return [False, u"Username too short"]
        for char in u"[ ,:;<>@()!\"'%&\\|\t\b":
            if char in username:
                return [False,
                        u"Username contains invalid character: '%s'" % char]

        # The localpart is acceptable if it can be parsed as the local part
        # of an RFC 2821 address.
        try:
            parseAddress("<%s@example.com>" % (username,))
        except ArgumentError:
            return [False, u"Username fails to parse"]

        # The domain is acceptable if it is one which we actually host.
        if domain not in self.getAvailableDomains():
            return [False, u"Domain not allowed"]

        query = self.store.query(userbase.LoginMethod,
                                 AND(userbase.LoginMethod.localpart == username,
                                     userbase.LoginMethod.domain == domain))
        return [not bool(query.count()), u"Username already taken"]


    def createUser(self, realName, username, domain, password, emailAddress):
        """
        Create a user, storing some associated metadata in the user's store,
        i.e. their first and last names (as a L{UserInfo} item), and a
        L{axiom.userbase.LoginMethod} allowing them to login with their email
        address.

        @param realName: the real name of the user.
        @type realName: C{unicode}

        @param username: the user's username.  they will be able to login with
            this.
        @type username: C{unicode}

        @param domain: the local domain - used internally to turn C{username}
            into a localpart@domain style string .
        @type domain: C{unicode}

        @param password: the password to be used for the user's account.
        @type password: C{unicode}

        @param emailAddress: the user's external email address.  they will be
            able to login with this also.
        @type emailAddress: C{unicode}

        @rtype: C{NoneType}
        """
        # XXX This method should be called in a transaction, it shouldn't
        # start a transaction itself.
        def _():
            loginsystem = self.store.findUnique(userbase.LoginSystem)

            # Create an account with the credentials they specified,
            # making it internal since it belongs to us.
            acct = loginsystem.addAccount(username, domain, password,
                                          verified=True, internal=True)

            # Create an external login method associated with the email
            # address they supplied, as well.  This creates an association
            # between that external address and their account object,
            # allowing password reset emails to be sent and letting them log
            # in to this account using that address as a username.
            emailPart, emailDomain = emailAddress.split("@")
            acct.addLoginMethod(emailPart, emailDomain, protocol=u"email",
                                verified=False, internal=False)
            substore = IBeneficiary(acct)
            # Record some of that signup information in case application
            # objects are interested in it.
            UserInfo(store=substore, realName=realName)
            self.product.installProductOn(substore)
        self.store.transact(_)

declareLegacyItem(typeName=UserInfoSignup.typeName,
                  schemaVersion=1,
                  attributes=dict(booth = reference(),
                                  benefactor = reference(),
                                  emailTemplate = text(),
                                  prompt = text(),
                                  prefixURL = text(allowNone=False)))

def userInfoSignup1To2(old):
    newProduct = old.store.findOrCreate(Product,
                                        types=list(
        chain(*[b.powerupNames for b in
                old.benefactor.benefactors('ascending')])))
    return old.upgradeVersion(UserInfoSignup.typeName, 1, 2,
                              booth=old.booth,
                              product=newProduct,
                              emailTemplate=old.emailTemplate,
                              prompt=old.prompt,
                              prefixURL=old.prefixURL)
upgrade.registerUpgrader(userInfoSignup1To2, UserInfoSignup.typeName, 1, 2)


class InitializerBenefactor(Item):
    typeName = 'initializer_benefactor'
    schemaVersion = 1

    realBenefactor = reference()

    def endow(self, ticket, beneficiary):
        beneficiary.findOrCreate(WebSite).installOn(beneficiary)
        beneficiary.findOrCreate(PrivateApplication).installOn(beneficiary)

        # They may have signed up in the past - if so, they already
        # have a password, and we should skip the initializer phase.
        substore = beneficiary.store.parent.getItemByID(beneficiary.store.idInParent)
        for acc in self.store.query(userbase.LoginAccount,
                                    userbase.LoginAccount.avatars == substore):
            if acc.password:
                self.realBenefactor.endow(ticket, beneficiary)
            else:
                beneficiary.findOrCreate(Initializer).installOn(beneficiary)
            break

    def resumeSignup(self, ticket, avatar):
        self.realBenefactor.endow(ticket, avatar)


class Initializer(Item):

    implements(INavigableElement)

    typeName = 'password_initializer'
    schemaVersion = 1

    installedOn = reference()

    powerupInterfaces = (INavigableElement,)

    def getTabs(self):
        # This won't ever actually show up
        return [Tab('Preferences', self.storeID, 1.0)]

    def setPassword(self, password):
        substore = self.store.parent.getItemByID(self.store.idInParent)
        for acc in self.store.parent.query(userbase.LoginAccount,
                                           userbase.LoginAccount.avatars == substore):
            acc.password = password
            self._delegateToBenefactor(acc)
            return

    def _delegateToBenefactor(self, loginAccount):
        site = self.store.parent
        ticket = site.findUnique(Ticket, Ticket.avatar == loginAccount)
        benefactor = ticket.benefactor
        benefactor.resumeSignup(ticket, self.store)

        self.store.powerDown(self, INavigableElement)
        self.deleteFromStore()



class InitializerPage(PublicPage):

    def __init__(self, original):
        for resource, domain in userbase.getAccountNames(original.installedOn):
            username = '%s@%s' % (resource, domain)
            break
        else:
            username = None
        PublicPage.__init__(self, original, original.store.parent, getLoader('initialize'),
                            IStaticShellContent(original.installedOn, None),
                            username)

    def render_head(self, ctx, data):
        tag = PublicPage.render_head(self, ctx, data)
        return tag[tags.script(src='/Mantissa/js/initialize.js')]

    def renderHTTP(self, ctx):
        req = inevow.IRequest(ctx)
        password = req.args.get('password', [None])[0]

        if password is None:
            return Page.renderHTTP(self, ctx)

        self.original.store.transact(self.original.setPassword,
                                     unicode(password)) # XXX TODO: select
                                                        # proper decoding
                                                        # strategy.
        return URL.fromString('/')

registerAdapter(InitializerPage,
                Initializer,
                inevow.IResource)



class Ticket(Item):
    schemaVersion = 2
    typeName = 'ticket'

    issuer = reference(allowNone=False)
    booth = reference(allowNone=False)
    avatar = reference()
    claimed = integer(default=0)
    product = reference(allowNone=False)

    email = text()
    nonce = text()

    def __init__(self, **kw):
        super(Ticket, self).__init__(**kw)
        self.booth.createdTicketCount += 1
        self.nonce = _generateNonce()

    def claim(self):
        if not self.claimed:
            log.msg("Claiming a ticket for the first time for %r" % (self.email,))
            username, domain = self.email.split('@', 1)
            realm = IRealm(self.store)
            acct = realm.accountByAddress(username, domain)
            if acct is None:
                acct = realm.addAccount(username, domain, None)
            self.avatar = acct
            self.claimed += 1
            self.booth.ticketClaimed(self)
            self.product.installProductOn(IBeneficiary(self.avatar))
        else:
            log.msg("Ignoring re-claim of ticket for: %r" % (self.email,))
        return self.avatar
    claim = transacted(claim)


def ticket1to2(old):
    """
    change Ticket to refer to Products and not benefactor factories.
    """
    if isinstance(old.benefactor, Multifactor):
        types = list(chain(*[b.powerupNames for b in
                old.benefactor.benefactors('ascending')]))
    elif isinstance(old.benefactor, InitializerBenefactor):
        #oh man what a mess
        types = list(chain(*[b.powerupNames for b in
                old.benefactor.realBenefactor.benefactors('ascending')]))
    newProduct = old.store.findOrCreate(Product,
                                        types=types)

    if old.issuer is None:
        issuer = old.store.findOrCreate(TicketBooth)
    else:
        issuer = old.issuer
    t = old.upgradeVersion(Ticket.typeName, 1, 2,
                           product = newProduct,
                           issuer = issuer,
                           booth = old.booth,
                           avatar = old.avatar,
                           claimed = old.claimed,

                           email = old.email,
                           nonce = old.nonce)

upgrade.registerUpgrader(ticket1to2, Ticket.typeName, 1, 2)
class _DelegatedBenefactor(Item):
    typeName = 'mantissa_delegated_benefactor'
    schemaVersion = 1

    benefactor = reference(allowNone=False)
    multifactor = reference(allowNone=False)
    order = integer(allowNone=False, indexed=True)



class Multifactor(Item):
    """
    A benefactor with no behavior of its own, but which collects
    references to other benefactors and delegates endowment
    responsibility to them.
    """

    typeName = 'mantissa_multi_benefactor'
    schemaVersion = 1

    order = integer(default=0)

    def benefactors(self, order):
        for deleg in self.store.query(_DelegatedBenefactor,
                                      _DelegatedBenefactor.multifactor == self,
                                      sort=getattr(_DelegatedBenefactor.order, order)):
            yield deleg.benefactor



class _SignupTracker(Item):
    """
    Signup-system private Item used to track which signup mechanisms
    have been created.
    """
    signupItem = reference()
    createdOn = timestamp()
    createdBy = text()



def _getPublicSignupInfo(siteStore):
    """
    Get information about public web-based signup mechanisms.

    @param siteStore: a store with some signups installed on it (as indicated
    by _SignupTracker instances).

    @return: a generator which yields 2-tuples of (prompt, url) where 'prompt'
    is unicode briefly describing the signup mechanism (e.g. "Sign Up"), and
    'url' is a (unicode) local URL linking to a page where an anonymous user
    can access it.
    """

    # Note the underscore; this _should_ be a public API but it is currently an
    # unfortunate hack; there should be a different powerup interface that
    # requires prompt and prefixURL attributes rather than _SignupTracker.
    # -glyph

    for tr in siteStore.query(_SignupTracker):
        si = tr.signupItem
        p = getattr(si, 'prompt', None)
        u = getattr(si, 'prefixURL', None)
        if p is not None and u is not None:
            yield (p, u'/'+u)


class SignupConfiguration(Item):
    """
    Provide administrative configuration tools for the signup options
    available on a Mantissa server.
    """
    typeName = 'mantissa_signup_configuration'
    schemaVersion = 1

    installedOn = reference()

    powerupInterfaces = (INavigableElement,)

    def getTabs(self):
        return [Tab('Admin', self.storeID, 0.5,
                    [Tab('Signup', self.storeID, 0.7)],
                    authoritative=False)]


    def getSignupSystems(self):
        return dict((p.name, p) for p in plugin.getPlugins(ISignupMechanism, plugins))

    def createSignup(self, creator, signupClass, signupConf,
                     product, emailTemplate, prompt):
        """
        Create a new signup facility in the site store's database.

        @param creator: a unicode string describing the creator of the new
        signup mechanism, for auditing purposes.

        @param signupClass: the item type of the signup mechanism to create.

        @param signupConf: a dictionary of keyword arguments for
        L{signupClass}'s constructor.

        @param product: A Product instance, describing the powerups to be
        installed with this signup.

        @param emailTemplate: a unicode string which contains some text that
        will be sent in confirmation emails generated by this signup mechanism
        (if any)

        @param prompt: a short unicode string describing this signup mechanism,
        as distinct from others.  For example: "Student Sign Up", or "Faculty
        Sign Up"

        @return: a newly-created, database-resident instance of signupClass.
        """

        siteStore = self.store.parent

        booth = siteStore.findOrCreate(TicketBooth, lambda booth: installOn(booth, siteStore))
        signupItem = signupClass(
            store=siteStore,
            booth=booth,
            product=product,
            emailTemplate=emailTemplate,
            prompt=prompt,
            **signupConf)
        siteStore.powerUp(signupItem)
        _SignupTracker(store=siteStore,
                       signupItem=signupItem,
                       createdOn=extime.Time(),
                       createdBy=creator)

        return signupItem

class ProductFormMixin(object):
    """
    Utility functions for rendering a form for choosing products to install.
    """
    def makeProductPicker(self):
        """
        Make a LiveForm with radio buttons for each Product in the store.
        """
        productPicker = liveform.LiveForm(
            self.coerceProduct,
            [liveform.Parameter(
              str(id(product)),
              liveform.FORM_INPUT,
              liveform.LiveForm(
              lambda selectedProduct, product=product: selectedProduct and product,
              [liveform.Parameter(
                'selectedProduct',
                liveform.RADIO_INPUT,
                bool,
                repr(product))]
              ))
              for product
              in self.original.store.parent.query(Product)],
            u"Product to Install")
        return productPicker

    def coerceProduct(self, **kw):
        """
        Convert the return value from the form to a list of Products.
        """
        return filter(None, kw.values())[0]




class SignupFragment(athena.LiveFragment, ProductFormMixin):
    fragmentName = 'signup-configuration'
    live = 'athena'

    def head(self):
        # i think this is the lesser evil.
        # alternatives being:
        #  * mangle form element names so we can put these in mantissa.css
        #    without interfering with similarly named things
        #  * put the following line of CSS into it's own file that is included
        #    by only this page
        #  * remove these styles entirely (makes the form unusable, the
        #    type="text" inputs are *tiny*)
        return tags.style(type='text/css')['''
        input[name=linktext], input[name=subject], textarea[name=blurb] { width: 40em }
        ''']

    def render_signupConfigurationForm(self, ctx, data):

        def makeSignupCoercer(signupPlugin):
            """
            Return a function that converts a selected flag and a set of
            keyword arguments into either None (if not selected) or a 2-tuple
            of (signupClass, kwargs).  signupClass is a callable which takes
            the kwargs as keyword arguments and returns an Item (a signup
            mechanism plugin gizmo).
            """
            def signupCoercer(selectedSignup, **signupConf):
                """
                Receive coerced values from the form post, massage them as
                described above.
                """
                if selectedSignup:
                    return signupPlugin.itemClass, signupConf
                return None
            return signupCoercer

        def coerceSignup(**kw):
            return filter(None, kw.values())[0]

        signupMechanismConfigurations = liveform.LiveForm(
            # makeSignupCoercer sets it up, we knock it down. (Nones returned
            # are ignored, there will be exactly one selected).
            coerceSignup,
            [liveform.Parameter(
                signupMechanism.name,
                liveform.FORM_INPUT,
                liveform.LiveForm(
                    makeSignupCoercer(signupMechanism),
                    [liveform.Parameter(
                        'selectedSignup',
                        liveform.RADIO_INPUT,
                        bool,
                        signupMechanism.description)] + signupMechanism.configuration,
                    signupMechanism.name))
             for signupMechanism
             in self.original.getSignupSystems().itervalues()],
            u"Signup Type")

        def coerceEmailTemplate(**k):
            return file(sibpath(__file__, 'signup.rfc2822')).read() % k

        emailTemplateConfiguration = liveform.LiveForm(
            coerceEmailTemplate,
            [liveform.Parameter('subject',
                                liveform.TEXT_INPUT,
                                unicode,
                                u'Email Subject',
                                default='Welcome to a Generic Axiom Application!'),
             liveform.Parameter('blurb',
                                liveform.TEXTAREA_INPUT,
                                unicode,
                                u'Blurb',
                                default=''),
             liveform.Parameter('linktext',
                                liveform.TEXT_INPUT,
                                unicode,
                                u'Link Text',
                                default="Click here to claim your 'generic axiom application' account")],
             description='Email Template')
        emailTemplateConfiguration.docFactory = getLoader('liveform-compact')

        existing = list(self.original.store.parent.query(_SignupTracker))
        if 0 < len(existing):
            deleteSignupForm = liveform.LiveForm(
                lambda **kw: self._deleteTrackers(k for (k, v) in kw.itervalues() if v),
                [liveform.Parameter('signup-' + str(i),
                                    liveform.CHECKBOX_INPUT,
                                    lambda wasSelected, tracker=tracker: (tracker, wasSelected),
                                    repr(tracker.signupItem))
                    for (i, tracker) in enumerate(existing)],
                description='Delete Existing Signups')
            deleteSignupForm.setFragmentParent(self)
        else:
            deleteSignupForm = ''

        productPicker = self.makeProductPicker()

        createSignupForm = liveform.LiveForm(
            self.createSignup,
            [liveform.Parameter('signupPrompt',
                                liveform.TEXT_INPUT,
                                unicode,
                                u'Descriptive, user-facing prompt for this signup',
                                default=u'Sign Up'),
             liveform.Parameter('product',
                                liveform.FORM_INPUT,
                                productPicker,
                                u'Pick some product'),
             liveform.Parameter('signupTuple',
                                liveform.FORM_INPUT,
                                signupMechanismConfigurations,
                                u'Pick just one dude'),
             liveform.Parameter('emailTemplate',
                                liveform.FORM_INPUT,
                                emailTemplateConfiguration,
                                u'You know you want to')],
            description='Create Signup')
        createSignupForm.setFragmentParent(self)

        return [deleteSignupForm, createSignupForm]


    def data_configuredSignupMechanisms(self, ctx, data):
        for _signupTracker in self.original.store.parent.query(_SignupTracker):
            yield {
                'typeName': _signupTracker.signupItem.__class__.__name__,
                'createdBy': _signupTracker.createdBy,
                'createdOn': _signupTracker.createdOn.asHumanly()}


    def createSignup(self,
                     signupPrompt,
                     signupTuple,
                     product,
                     emailTemplate):
        """

        @param signupPrompt: a short unicode string describing this new signup
        mechanism to disambiguate it from others.  For example: "sign up".

        @param signupTuple: a 2-tuple of (signupMechanism, signupConfig),

        """
        (signupMechanism, signupConfig) = signupTuple
        t = self.original.store.transact
        t(self.original.createSignup,
          self.page.username,
          signupMechanism,
          signupConfig,
          product,
          emailTemplate,
          signupPrompt)
        return u'Great job.'
    expose(createSignup)

    def _deleteTrackers(self, trackers):
        """
        Delete the given signup trackers and their associated signup resources.

        @param trackers: sequence of L{_SignupTrackers}
        """

        for tracker in trackers:
            if tracker.store is None:
                # we're not updating the list of live signups client side, so
                # we might get a signup that has already been deleted
                continue

            sig = tracker.signupItem

            # XXX the only reason we're doing this here is that we're afraid to
            # add a whenDeleted=CASCADE to powerups because it's inefficient,
            # however, this is arguably the archetypical use of
            # whenDeleted=CASCADE.  Soon we need to figure out a real solution
            # (but I have no idea what it is). -glyph

            for iface in sig.store.interfacesFor(sig):
                sig.store.powerDown(sig, iface)
            tracker.deleteFromStore()
            sig.deleteFromStore()


registerAdapter(SignupFragment, SignupConfiguration, INavigableFragment)




"""
Generate a stub for the tests for the deletion of the RealName item.
"""

from axiom.test.historic.stubloader import saveStub

from xmantissa.people import Person, RealName


def createDatabase(store):
    """
    Make a L{Person} with a corresponding L{RealName} which will be deleted.
    """
    person = Person(store=store)
    name = RealName(store=store, person=person)


if __name__ == '__main__':
    saveStub(createDatabase, 13508)



"""
Tests for L{Mugshot}'s version 2 to version 3 upgrader.
"""
from twisted.trial.unittest import SkipTest

from axiom.test.historic.stubloader import StubbedTest

from xmantissa.people import Mugshot, Person
from xmantissa.test.historic.stub_mugshot2to3 import (
    MUGSHOT_TYPE, MUGSHOT_BODY_PATH_SEGMENTS)


class MugshotUpgraderTestCase(StubbedTest):
    """
    Tests for L{Mugshot}'s version 2 to version 3 upgrader.
    """
    def setUp(self):
        """
        Skip the tests if PIL is unavailable.
        """
        try:
            import PIL
        except ImportError:
            raise SkipTest('PIL is not available')
        return StubbedTest.setUp(self)


    def test_attributesCopied(self):
        """
        The C{person}, C{smallerBody} and C{type} attributes of L{Mugshot}
        should have been copied over from the previous version.
        """
        from PIL import Image
        mugshot = self.store.findUnique(Mugshot)
        self.assertIdentical(mugshot.person, self.store.findUnique(Person))
        self.assertEqual(mugshot.type, MUGSHOT_TYPE)
        self.assertEqual(
            mugshot.body, self.store.newFilePath(*MUGSHOT_BODY_PATH_SEGMENTS))
        # mugshot.body should be untouched, it should have the same dimensions
        # as test/resources/square.png (240x240)
        self.assertEqual(Image.open(mugshot.body.open()).size, (240, 240))


    def test_smallerBodyAttribute(self):
        """
        L{Mugshot.smallerBody} should point to an image with the same
        dimensions as the current value of L{Mugshot.smallerSize}.
        """
        from PIL import Image
        mugshot = self.store.findUnique(Mugshot)
        self.assertEqual(
            Image.open(mugshot.smallerBody.open()).size,
            (mugshot.smallerSize, mugshot.smallerSize))




"""

Tests for inter-store messaging module, L{xmantissa.messaging}.

This module contains tests for persistent messaging between different accounts.

"""

import gc
from datetime import timedelta

from zope.interface import implements

from twisted.trial.unittest import TestCase
from twisted.internet.defer import Deferred

from twisted.protocols.amp import Box, Command, Integer, String

from epsilon.extime import Time

from axiom.iaxiom import IScheduler
from axiom.store import Store
from axiom.errors import UnsatisfiedRequirement
from axiom.item import Item, POWERUP_BEFORE
from axiom.attributes import text, bytes, integer, boolean, inmemory

from axiom.userbase import LoginSystem, LoginMethod, LoginAccount

from axiom.dependency import installOn

from axiom.scheduler import TimedEvent

from xmantissa.interstore import (
    # Public Names
    MessageQueue, AMPMessenger, LocalMessageRouter, Value,
    AMPReceiver, commandMethod, answerMethod, errorMethod,
    SenderArgument, TargetArgument,

    # Constants
    AMP_MESSAGE_TYPE, AMP_ANSWER_TYPE, DELIVERY_ERROR,

    # Error Types
    ERROR_REMOTE_EXCEPTION, ERROR_NO_SHARE, ERROR_NO_USER, ERROR_BAD_SENDER,

    # Private Names
    _RETRANSMIT_DELAY, _QueuedMessage, _AlreadyAnswered, _FailedAnswer,
    _AMPExposer, _AMPErrorExposer)

from xmantissa.sharing import getEveryoneRole, Identifier
from xmantissa.error import (
    MessageTransportError, BadSender, UnknownMessageType, RevertAndRespond,
    MalformedMessage)
from xmantissa.ixmantissa import IMessageReceiver, IMessageRouter

class SampleException(Exception):
    """
    Something didn't happen because of a problem.
    """



class StubReceiver(Item):
    """
    This is a message receiver that will store a message sent to it for
    inspection by tests.
    """
    implements(IMessageReceiver)

    messageType = text(
        doc="""
        The message type which C{messageReceived} should put into its return
        value.
        """)
    messageData = bytes(
        doc="""
        The message data which C{messageReceived} should put into its return
        value.
        """)
    inconsistent = boolean(
        doc="""
        This value is set to True during the execution of C{messageReceived},
        but False afterwards.  If everything is properly transactional it
        should never be observably false by other code.
        """)

    buggy = boolean(allowNone=False, default=False,
                    doc="""
                    C{messageReceived} should raise a L{SampleException}.
                    """)

    badReturn = boolean(allowNone=False, default=False,
                        doc="""
                        C{messageReceived} should return L{None}.
                        """)

    receivedCount = integer(default=0,
                            doc="""
                            This is a counter of the number of messages
                            received by C{messageReceived}.
                            """)

    reciprocate = boolean(allowNone=False, default=False,
                          doc="""
                          C{messageReceived} should respond to its C{sender}
                          parameter with a symmetric message in addition to
                          answering.
                          """)

    revertType = text(allowNone=True,
                      doc="""
                      If set, this specifies the type of the
                      L{RevertAndRespond} exception that C{messageReceived}
                      should raise.
                      """)

    revertData = bytes(allowNone=True,
                       doc="""
                       If C{revertType} is set, this specifies the data of the
                       L{RevertAndRespond} exception that C{messageReceived}
                       should raise.
                       """)


    def messageQueue(self):
        """
        This is a temporary workaround; see ticket #2640 for details on the way
        this method should be implemented in the future.
        """
        return self.store.findUnique(MessageQueue)


    def messageReceived(self, value, sender, receiver):
        """
        A message was received.  Increase the message counter and store its
        contents.
        """
        self.receivedCount += 1
        self.messageType = value.type
        self.messageData = value.data
        self.inconsistent = True
        if self.buggy:
            raise SampleException("Sample Message")
        if self.revertType is not None:
            raise RevertAndRespond(Value(self.revertType,
                                            self.revertData))
        self.inconsistent = False
        if self.badReturn:
            return None
        if self.reciprocate:
            self.messageQueue().queueMessage(
                receiver, sender, Value(value.type + u'.response',
                                           value.data + ' response'))
        return Value(u"custom.message.type", "canned response")



class StubSlowRouter(Item):
    """
    Like L{LocalMessageRouter}, but don't actually deliver the messages until
    the test forces them to be delivered.

    By way of several parameters to `flushMessages`, this stub implementation
    allows for all of the arbitrary ways in which a potential networked
    implementation is allowed to behave - dropping messages, repeating
    messages, and even failing in buggy ways.

    Note: this must be kept in memory for the duration of any test using it.

    @ivar messages: a list of (sender, target, value, messageID) tuples
    received by routeMessage.

    @ivar acks: a list of (deferred, (sender, target, value, messageID))
    tuples, representing an answer received by routeAnswer and the deferred
    that was returned to indicate its delivery.
    """

    dummy = integer(
        doc="""
        No state on this item is persistent; this is just to satisfy Axiom's schema
        requirement.
        """)

    messages = inmemory()
    acks = inmemory()

    def localRouter(self):
        """
        Return a L{LocalMessageRouter} for this slow router's store.
        """
        return LocalMessageRouter(self.store.findUnique(LoginSystem))


    def activate(self):
        """
        Initialize temporary list to queue messages.
        """
        self.messages = []
        self.acks = []


    def routeMessage(self, sender, target, value, messageID):
        """
        Stub implementation of L{IMessageRouter.routeMessage} that just appends
        to a list in memory, and later delegates from that list to the local
        router.
        """
        self.messages.append((sender, target, value, messageID))


    def routeAnswer(self, originalSender, originalTarget, value, messageID):
        """
        Stub implementation of L{IMessageRouter.routeAnswer} that just
        appends to a list in memory.
        """
        D = Deferred()
        self.acks.append((D, (originalSender, originalTarget, value,
                              messageID)))
        return D


    def flushMessages(self, dropAcks=False,
                      dropAckErrorType=MessageTransportError,
                      stallAcks=False,
                      repeatAcks=False):
        """
        Delegate all messages queued in memory with routeMessage to the
        specified local router.

        @param dropAcks: a boolean, indicating whether to drop the answers
        queued by routeAnswer.

        @param dropAckErrorType: an exception type, indicating what exception
        to errback the Deferreds returned by routeAnswer with.

        @param stallAcks: a boolean, indicating whether to keep, but not act,
        on the answers queued by routeAnswer.

        @param repeatAcks: a boolean, indicating whether to repeat all of the
        acks the next time flushMessages is called.
        """
        m = self.messages[:]
        self.messages = []
        for message in m:
            self.localRouter().routeMessage(*message)
        if dropAcks:
            for D, ack in self.acks:
                D.errback(dropAckErrorType())
            self.acks = []
        if not stallAcks:
            for D, ack in self.acks:
                self.localRouter().routeAnswer(*ack).chainDeferred(D)
            if repeatAcks:
                # the Deferreds are used up, so we need a fresh batch for the
                # next run-through (although these will be ignored)
                self.acks = [(Deferred(), ack) for (D, ack) in self.acks]
            else:
                self.acks = []


    def spuriousDeliveries(self):
        """
        Simulate a faulty transport, and deliver all the currently pending
        messages without paying attention to their results.
        """
        for message in self.messages:
            self.localRouter().routeMessage(*message)



class StubDeliveryConsequence(Item):
    """
    This implements a delivery consequence.

    @ivar responses: a tuple of (answer-type, answer-data, message-type,
    message-data, sender, target), listing all the answers received by
    answerReceived.

    @ivar bucket: a list which will have this L{StubDeliveryConsequence}
    appended to it when a successful message is processed.
    """

    responses = inmemory()
    bucket = inmemory()

    invocations = integer(
        """
        Counter, keeping track of how many times this consequence has been
        invoked.
        """,
        default=0, allowNone=False)

    succeeded = boolean(
        """
        Did the action succeed?  None if it hasn't completed, True if yes,
        False if no.
        """)

    inconsistent = boolean(
        """
        This should never be set to True.  It's set to None by default, False
        when the callback fully succeeds.
        """)

    buggy = boolean(
        """
        Set this to cause 'success' to raise an exception.
        """,
        default=False,
        allowNone=False)

    def activate(self):
        """
        Initialize the list of received responses.
        """
        self.responses = []
        self.bucket = []


    def success(self):
        """
        A response was received to the message.  This will be executed in a
        transaction.  Raise an exception if this consequence is buggy.
        """
        self.bucket.append(self)
        self.inconsistent = True
        self.invocations += 1
        self.succeeded = True
        if self.buggy:
            raise SampleException()
        self.inconsistent = False


    def failure(self):
        """
        The message could not be delivered for some reason.  This will be
        executed in a transaction.  Raise an exception if this consequence is
        buggy.

        @param reason: an exception.
        """
        self.invocations += 1
        self.succeeded = False


    def answerReceived(self, answerValue, originalValue,
                       originalSender, originalTarget):
        """
        An answer was received.
        """
        if answerValue.type == DELIVERY_ERROR:
            self.failure()
        else:
            self.success()
        # It's important that this happen after the "application" logic so that
        # the tests will not see this set if an exception has been raised.
        self.responses.append((answerValue.type, answerValue.data,
                               originalValue.type, originalValue.data,
                               originalSender, originalTarget))



class TimeFactory(object):
    """
    Make a fake time factory.
    """

    def __init__(self):
        """
        Create a time factory with some default values.
        """
        self.currentSeconds = 0.0


    def advance(self):
        """
        Advance the current time by one second.
        """
        self.currentSeconds += 1.0


    def next(self):
        """
        Produce the next time in the sequence, then advance.
        """
        self.advance()
        return Time.fromPOSIXTimestamp(self.currentSeconds)


    def peek(self):
        """
        Return the value that will come from the next call to 'next'.
        """
        return Time.fromPOSIXTimestamp(self.currentSeconds + 1)



class SingleSiteMessagingTests(TestCase):
    """
    These are tests for messaging within a single configured site store.
    """

    def setUp(self):
        """
        Create a site store with two users that can send messages to each
        other.
        """
        self.siteStore = Store()
        self.time = TimeFactory()
        self.loginSystem = LoginSystem(store=self.siteStore)
        installOn(self.loginSystem, self.siteStore)

        self.aliceAccount = self.loginSystem.addAccount(
            u"alice", u"example.com", u"asdf", internal=True)
        self.bobAccount = self.loginSystem.addAccount(
            u"bob", u"example.com", u"asdf", internal=True)

        self.aliceStore, self.aliceQueue = self.accountify(
            self.aliceAccount.avatars.open())
        self.bobStore, self.bobQueue = self.accountify(
            self.bobAccount.avatars.open())

        # I need to make a target object with a message receiver installed on
        # it.  Then I need to share that object.
        self.receiver = StubReceiver(store=self.bobStore)
        getEveryoneRole(self.bobStore).shareItem(self.receiver, u"suitcase")

        self.retransmitDelta = timedelta(seconds=_RETRANSMIT_DELAY)


    def accountify(self, userStore):
        """
        Add a MessageQueue to the given user store and stub out its scheduler's
        time function.
        """
        queue = MessageQueue(store=userStore)
        installOn(queue, userStore)
        IScheduler(userStore).now = self.time.peek
        return userStore, queue


    def runQueue(self, queue):
        """
        Advance the current time and run the given message queue.
        """
        self.time.advance()
        return queue.run()


    def test_bogusConfiguration(self):
        """
        Delivering a message on a site without a L{LoginSystem} should cause an
        L{UnsatisfiedRequirement} exception to be logged, and the message not
        to be delivered.
        """
        self.loginSystem.deleteFromStore()
        self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        [err] = self.flushLoggedErrors(UnsatisfiedRequirement)
        # The message should still be queued.
        self.assertEqual(len(list(self.aliceStore.query(_QueuedMessage))),
                         1)


    def unattachedUserStore(self):
        """
        Create a store that is structured as if it was a user-store with
        messaging enabled, but was opened un-attached from its parent store.

        @return: a 2-tuple of (store, queue).
        """
        carolStore, carolQueue = self.accountify(Store())
        acct = LoginAccount(store=carolStore, password=u'asdf')
        lm = LoginMethod(store=carolStore, localpart=u'carol',
                         domain=u'example.com', internal=True, protocol=u'*',
                         account=acct, verified=True)
        return carolStore, carolQueue


    def test_accidentallyOpenAsSite(self):
        """
        If a message delivery is somehow attempted with a user store
        accidentally opened as a site store, delivery should fail.

        Normally this will not happen, since the current implementation (as of
        when this test was written) of the scheduler will not allow timed
        events to run with a L{SubScheduler} installed rather than a
        L{Scheduler}.  However, eliminating this distinction is a long-term
        goal, so this test is a defense against both future modifications and
        other code which may emulate scheduler APIs.
        """
        carolStore, carolQueue = self.unattachedUserStore()
        sdc = StubDeliveryConsequence(store=carolStore)
        carolQueue.queueMessage(
            Identifier(u"nothing", u"carol", u"example.com"),
            Identifier(u"suitcase", u"bob", u"example.com"),
            Value(u'custom.message.type', "Some message contents"), sdc)
        self.runQueue(carolQueue)
        [err] = self.flushLoggedErrors(UnsatisfiedRequirement)
        # The message should still be queued.
        self.assertEqual(len(list(carolStore.query(_QueuedMessage))), 1)


    def test_accidentallyAnswerAsSite(self):
        """
        If an answer delivery is somehow attempted with a user store
        accidentally opened as a site store, the delivery should result in a
        transient failure.

        This is even less likely than the case described in
        L{test_accidentallyOpenAsSite}, but in the unlikely event that the
        scheduler is manually run, it still shouldn't result in any errors
        being logged or state being lost.
        """
        carolStore, carolQueue = self.unattachedUserStore()
        carolReceiver = StubReceiver(store=carolStore)
        getEveryoneRole(carolStore).shareItem(carolReceiver, u'some-share-id')
        bogusID = Identifier(u'nothing', u'nobody', u'nowhere')
        carolID = Identifier(u'some-share-id', u'carol', u'example.com')
        carolQueue.routeMessage(bogusID, carolID, Value(u'no.type', 'contents'), 1)
        [err] = self.flushLoggedErrors(UnsatisfiedRequirement)
        # The answer should still be queued.
        self.assertEqual(len(list(carolStore.query(_AlreadyAnswered))), 1)


    def test_queueMessageSimple(self):
        """
        Queuing a message should create a _QueuedMessage object and schedule it
        for delivery to its intended recipient.
        """

        # Maybe I should do this by sharing an object in Alice's store and then
        # wrapping a SharingView-type thing around it?  it seems like there
        # ought to be a purely model-level API for this, though.
        self.aliceQueue.queueMessage(
            Identifier(u"nothing", u"alice", u"example.com"),
            Identifier(u"suitcase", u"bob", u"example.com"),
            Value(u"custom.message.type", "This is an important message."))

        # OK now let's find the message that was queued.
        qm = self.aliceStore.findUnique(_QueuedMessage)
        self.assertEqual(qm.senderUsername, u'alice')
        self.assertEqual(qm.senderDomain, u'example.com')

        # Can you do this?  It seems like it might be inconvenient to always
        # determine a resolvable "return address" - the test case here is a
        # simulation of reasonable behavior; alice hasn't shared anything.
        # self.assertEqual(qm.senderShareID, None)
        self.assertEqual(qm.targetUsername, u"bob")
        self.assertEqual(qm.targetDomain, u"example.com")
        self.assertEqual(qm.targetShareID, u"suitcase")
        self.assertEqual(qm.value.type, u"custom.message.type")
        self.assertEqual(qm.value.data, "This is an important message.")

        # It should be scheduled.  Is there a timed event?
        te = self.aliceStore.findUnique(
            TimedEvent,
            TimedEvent.runnable == self.aliceQueue)

        # It should be scheduled immediately.  This uses the real clock, but in
        # a predictable way (i.e. if time does not go backwards, then this will
        # work).  If this test ever fails intermittently there _is_ a real
        # problem.
        self.assertNotIdentical(te.time, None)
        self.failUnless(te.time <= Time())
        runresult = self.runQueue(self.aliceQueue)

        # It should succeed, it should not reschedule itself; the scheduler
        # will delete things that return None from run().  It would be nice to
        # integrate with the scheduler here, but that would potentially drag in
        # dependencies on other systems not scheduling stuff.
        self.assertEqual(runresult, None)

        self.assertEqual(self.receiver.messageData,
                         "This is an important message.")
        self.assertEqual(self.receiver.messageType, u"custom.message.type")
        self.assertEqual(list(self.aliceStore.query(_QueuedMessage)), [])


    def aliceToBobWithConsequence(self, buggy=False):
        """
        Queue a message from Alice to Bob with a supplied
        L{StubDeliveryConsequence} and return it.
        """
        sdc = StubDeliveryConsequence(store=self.aliceStore,
                                      buggy=buggy)
        self.aliceQueue.queueMessage(
            Identifier(u"nothing", u"alice", u"example.com"),
            Identifier(u"suitcase", u"bob", u"example.com"),
            Value(u'custom.message.type', "Some message contents"), sdc)
        return sdc


    def checkOneResponse(self, sdc,
                         expectedType=u"custom.message.type",
                         expectedData="canned response",
                         originalSender=
                         Identifier(u"nothing", u"alice", u"example.com"),
                         originalTarget=
                         Identifier(u"suitcase", u"bob", u"example.com"),
                         succeeded=None):
        """
        This checks that the response received has the expected type and data,
        and corresponds to the sender and target specified by
        L{SingleSiteMessagingTests.aliceToBobWithConsequence}.
        """
        if succeeded is None:
            if expectedType == DELIVERY_ERROR:
                succeeded = False
            else:
                succeeded = True
        # First, let's make sure that transaction committed.
        self.assertEqual(sdc.succeeded, succeeded)
        self.assertEqual(sdc.responses,
                         [(expectedType, expectedData,
                           u'custom.message.type', # type
                           "Some message contents", # data
                           originalSender, originalTarget
                           )])


    def test_queueMessageSuccessNotification(self):
        """
        Queueing a message should emit a success notification to the supplied
        'consequence' object.
        """
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        self.assertEqual(sdc.succeeded, True)
        self.checkOneResponse(sdc)


    def test_queueMessageSuccessErrorHandling(self):
        """
        If the supplied 'consequence' object is buggy, the error should be
        logged so that the answer can be processed later, but not propagated to
        the network layer.
        """
        sdc = self.aliceToBobWithConsequence(True)
        self.runQueue(self.aliceQueue)
        # It should be run in a transaction so none of the stuff set by
        # 'success' should be set
        self.assertEqual(sdc.succeeded, None)
        self.assertEqual(sdc.inconsistent, None)
        self.assertEqual(sdc.invocations, 0)
        [err] = self.flushLoggedErrors(SampleException)

        # Make sure that no messages are queued.
        self.assertEqual(list(self.aliceStore.query(_QueuedMessage)), [])
        failures = list(self.aliceStore.query(_FailedAnswer))
        self.assertEqual(len(failures), 1)

        # Fix the bug.  In normal operation this would require a code upgrade.
        sdc.buggy = False
        failures[0].redeliver()
        self.checkOneResponse(sdc)


    def test_alreadyAnsweredRemoval(self):
        """
        L{_AlreadyAnswered} records should be removed after the deferred from
        L{routeAnswer} is fired.
        """
        slowRouter = self.stubSlowRouter()
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages(stallAcks=True)
        self.assertEqual(len(slowRouter.acks), 1)
        # sanity check
        self.assertEqual(self.bobStore.query(_AlreadyAnswered).count(), 1)
        slowRouter.flushMessages()
        self.assertEqual(self.bobStore.query(_AlreadyAnswered).count(), 0)


    def test_repeatedAnswer(self):
        """
        If answers are repeated, they should only be processed once.
        """
        slowRouter = self.stubSlowRouter()
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages(repeatAcks=True)
        slowRouter.flushMessages()
        self.assertEqual(sdc.invocations, 1)


    def _reschedulingTest(self, errorType):
        """
        Test for rescheduling of L{_AlreadyAnswered} results in the presence of
        the given error from the router.
        """
        slowRouter = self.stubSlowRouter()
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages(dropAcks=True, dropAckErrorType=errorType)
        # It should be scheduled.
        self.assertEqual(
            len(list(IScheduler(self.bobQueue.store).scheduledTimes(self.bobQueue))),
            1)
        # Now let's run it and see if the ack gets redelivered.
        self.runQueue(self.bobQueue)
        slowRouter.flushMessages()
        self.assertEqual(sdc.succeeded, True)


    def test_alreadyAnsweredReschedule(self):
        """
        L{_AlreadyAnswered} records should be scheduled for retransmission if
        the L{Deferred} from L{routeAnswer} is errbacked with a
        L{MessageTransportError}.  No message should be logged, since this is a
        transient and potentially expected error.
        """
        self._reschedulingTest(MessageTransportError)


    def test_alreadyAnsweredRescheduleAndLog(self):
        """
        L{_AlreadyAnswered} records should be scheduled for retransmission if
        the L{Deferred} from L{routeAnswer} is errbacked with an unknown
        exception type, and the exception should be logged.
        """
        self._reschedulingTest(SampleException)
        [err] = self.flushLoggedErrors(SampleException)


    def test_alreadyAnsweredRescheduleCrash(self):
        """
        L{_AlreadyAnswered} records should be scheduled for retransmission if
        the L{Deferred} from L{routeAnswer} dies without being callbacked or
        errbacked (such as if the store were to crash).
        """
        slowRouter = self.stubSlowRouter()
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages(stallAcks=True)
        self.assertEqual(sdc.invocations, 0)
        slowRouter.acks = []
        gc.collect()
        # Make sure the Deferred is well and truly gone.
        self.assertIdentical(
            self.bobStore.findUnique(_AlreadyAnswered).deliveryDeferred,
            None)
        self.runQueue(self.bobQueue)
        slowRouter.flushMessages()
        self.assertEqual(sdc.invocations, 1)


    def test_noRemoteUser(self):
        """
        What if the target user we're trying to talk to doesn't actually exist
        in the system?  The message delivery should fail.
        """
        sdc = StubDeliveryConsequence(store=self.aliceStore)
        self.aliceQueue.queueMessage(
            Identifier(u"nothing", u"alice", u"example.com"),
            Identifier(u"suitcase", u"bohb", u"example.com"),
            Value(u"custom.message.type", "Some message contents"), sdc)
        self.runQueue(self.aliceQueue)
        self.assertEqual(sdc.succeeded, False)
        self.checkOneResponse(
            sdc, DELIVERY_ERROR, ERROR_NO_USER,
            originalTarget=Identifier(u"suitcase", u"bohb", u"example.com"))
        self.assertEqual(sdc.invocations, 1)


    def test_noRemoteShare(self):
        """
        Similarly, if there's nothing identified by the shareID specified, the
        message delivery should fail.
        """
        sdc = StubDeliveryConsequence(store=self.aliceStore)
        self.aliceQueue.queueMessage(
            Identifier(u"nothing", u"alice", u"example.com"),
            Identifier(u"nothing", u"bob", u"example.com"),
            Value(u"custom.message.type", "Some message contents"), sdc)
        self.runQueue(self.aliceQueue)
        self.assertEqual(sdc.succeeded, False)
        self.checkOneResponse(
            sdc, DELIVERY_ERROR, ERROR_NO_SHARE,
            originalTarget=Identifier(u"nothing", u"bob", u"example.com"))
        self.assertEqual(sdc.invocations, 1)


    def buggyReceiverTest(self, exceptionType):
        """
        Run a test expecting the receiver to fail.
        """
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        self.assertEqual(sdc.succeeded, False)
        self.checkOneResponse(sdc, DELIVERY_ERROR, ERROR_REMOTE_EXCEPTION)
        [err] = self.flushLoggedErrors(exceptionType)
        self.assertEqual(sdc.invocations, 1)
        self.assertEqual(self.receiver.inconsistent, None)


    def test_messageReceivedBadReturn(self):
        """
        When L{messageReceived} does not properly return a 2-tuple, that
        resulting exception should be reported to the delivery consequence of
        the message.  The target database should not be left in an inconsistent
        state.
        """
        self.receiver.badReturn = True
        self.buggyReceiverTest(TypeError)


    def test_messageReceivedException(self):
        """
        When L{messageReceived} raises an exception, that exception should be
        reported to the delivery consequence of the message.  The target
        database should not be left in an inconsistent state.
        """
        self.receiver.buggy = True
        self.buggyReceiverTest(SampleException)


    def test_revertAndRespond(self):
        """
        When L{messageReceived} raises the special L{RevertAndRespond}
        exception, the values passed to the exception should be used to
        generate the response, but the transaction should be reverted.
        """
        t = self.receiver.revertType = u'custom.reverted.type'
        d = self.receiver.revertData = "this is some data that I reverted"
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        self.assertEqual(sdc.succeeded, True)
        self.assertEqual(sdc.inconsistent, False)
        self.assertEqual(sdc.invocations, 1)
        self.assertEqual(self.receiver.inconsistent, None)
        self.checkOneResponse(sdc, t, d)


    def test_droppedException(self):
        """
        When L{messageReceived} raises an exception, that exception should be
        reported to the delivery consequence of the message, even if the
        initial transmission of the error report is lost.
        """
        slowRouter = self.stubSlowRouter()
        self.receiver.buggy = True
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages(dropAcks=True)
        [err] = self.flushLoggedErrors(SampleException)
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages()
        self.assertEqual(sdc.invocations, 1)
        self.assertEqual(sdc.succeeded, False)
        self.checkOneResponse(sdc, DELIVERY_ERROR, ERROR_REMOTE_EXCEPTION)


    def test_senderNotVerified(self):
        """
        When the sender users name or domain do not match an internal, verified
        login method of the originating store, sending a message via
        queueMessage should resport an ERROR_BAD_SENDER.
        """
        sdc = StubDeliveryConsequence(store=self.aliceStore)
        self.aliceQueue.queueMessage(
            Identifier(u"nothing", u"fred", u"example.com"),
            Identifier(u"suitcase", u"bob", u"example.com"),
            Value(u"custom.message.type", "Some message contents"),
            sdc)
        self.assertEqual(sdc.invocations, 0)
        self.runQueue(self.aliceQueue)
        self.assertEqual(self.receiver.receivedCount, 0)
        self.assertEqual(sdc.succeeded, False)
        self.assertEqual(sdc.invocations, 1)
        self.checkOneResponse(sdc, DELIVERY_ERROR, ERROR_BAD_SENDER,
                              originalSender=Identifier(
                u"nothing", u"fred", u"example.com"))
        [err] = self.flushLoggedErrors(BadSender)
        bs = err.value
        self.assertEqual(bs.attemptedSender, u'fred@example.com')
        self.assertEqual(bs.allowedSenders, [u'alice@example.com'])
        self.assertEqual(str(bs),
                         "alice@example.com attempted to send message "
                         "as fred@example.com")


    def stubSlowRouter(self):
        """
        Replace this test's stub router with an artificially slowed-down
        router.
        """
        slowRouter = StubSlowRouter(store=self.siteStore)
        self.siteStore.powerUp(slowRouter, IMessageRouter, POWERUP_BEFORE)
        return slowRouter


    def test_slowDelivery(self):
        """
        If the site message-deliverer powerup returns a Deferred that takes a
        while to fire, the L{MessageQueue} should respond by rescheduling
        itself in the future.
        """
        slowRouter = self.stubSlowRouter()
        self.aliceQueue.queueMessage(
            Identifier(u"nothing", u"alice", u"example.com"),
            Identifier(u"suitcase", u"bob", u"example.com"),
            Value(u"custom.message.type", "Message2"))
        [time1] = IScheduler(self.aliceQueue.store).scheduledTimes(self.aliceQueue)
        time2 = self.runQueue(self.aliceQueue)
        self.assertEqual(time2 - self.time.peek(), self.retransmitDelta)
        self.assertEqual(self.receiver.receivedCount, 0)
        self.assertEqual(len(slowRouter.messages), 1)
        slowRouter.flushMessages()
        self.assertEqual(len(slowRouter.messages), 0)
        self.assertEqual(self.receiver.receivedCount, 1)


    def test_reallySlowDelivery(self):
        """
        If the Deferred takes so long to fire that another retransmission
        attempt takes place, the message should only be delivered once.  If it
        does fail, the next transmission attempt should actually transmit.
        """
        slowRouter = self.stubSlowRouter()
        c = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages()
        self.assertEqual(len(slowRouter.messages), 0)
        self.assertEqual(self.receiver.receivedCount, 1)
        self.assertEqual(c.invocations, 1)
        self.assertEqual(c.succeeded, 1)


    def test_multipleMessages(self):
        """
        Sending multiple messages at the same time should result in the
        messages being processed immediately, with no delay, but in order.
        """
        slowRouter = self.stubSlowRouter()
        bucket = []
        c = self.aliceToBobWithConsequence()
        c.bucket = bucket
        c2 = self.aliceToBobWithConsequence()
        c2.bucket = bucket
        # Sanity check; make sure the message hasn't been processed yet.
        self.assertEqual(bucket, [])
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages()
        self.assertEqual(c.invocations, 1)
        self.assertEqual(c.succeeded, 1)
        self.assertEqual(c2.invocations, 1)
        self.assertEqual(c2.succeeded, 1)
        self.assertEqual(bucket, [c, c2])
        self.assertEqual(self.runQueue(self.aliceQueue), None)


    def test_multipleAnswers(self):
        """
        Sending multiple messages at the same time should result in the
        messages being processed immediately, with no delay, but in order.
        """
        slowRouter = self.stubSlowRouter()
        bucket = []
        c = self.aliceToBobWithConsequence()
        c.bucket = bucket
        c2 = self.aliceToBobWithConsequence()
        c2.bucket = bucket
        # Sanity check; make sure the message hasn't been processed yet.
        self.assertEqual(bucket, [])
        self.runQueue(self.aliceQueue)
        slowRouter.flushMessages(dropAcks=True)
        [time1] = IScheduler(self.bobQueue.store).scheduledTimes(self.bobQueue)
        time2 = self.runQueue(self.bobQueue)
        self.assertEqual(time2 - self.time.peek(), self.retransmitDelta)
        slowRouter.flushMessages()
        self.assertEqual(c.invocations, 1)
        self.assertEqual(c.succeeded, 1)
        self.assertEqual(c2.invocations, 1)
        self.assertEqual(c2.succeeded, 1)
        self.assertEqual(bucket, [c, c2])
        self.assertEqual(self.runQueue(self.aliceQueue), None)


    def test_deliveryIdempotence(self):
        """
        Delivering the same message to a substore twice should only result in
        it being delivered to application code once.
        """
        slowRouter = self.stubSlowRouter()
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        slowRouter.spuriousDeliveries()
        self.assertEqual(self.receiver.receivedCount, 1)
        self.assertEqual(sdc.invocations, 0)
        slowRouter.flushMessages()
        self.assertEqual(self.receiver.receivedCount, 1)
        self.assertEqual(sdc.invocations, 1)
        self.checkOneResponse(sdc, )


    def test_reciprocate(self):
        """
        In addition to responding to the message with a return value, the
        receiver should be able to remember the sender and emit reciprocal
        messages.
        """
        self.receiver.reciprocate = True
        receiver2 = StubReceiver(store=self.aliceStore)
        getEveryoneRole(self.aliceStore).shareItem(receiver2, u'nothing')
        sdc = self.aliceToBobWithConsequence()
        self.runQueue(self.aliceQueue)
        self.runQueue(self.bobQueue)
        self.assertEqual(receiver2.receivedCount, 1)
        self.assertEqual(receiver2.messageType, u'custom.message.type.response')
        self.assertEqual(receiver2.messageData,
                         'Some message contents response')


    def test_unhandledDeliveryError(self):
        """
        When a message cannot be delivered, but no consequence is supplied, the
        error should be logged.
        """
        self.receiver.buggy = True
        self.aliceQueue.queueMessage(
            Identifier(u"nothing", u"alice", u"example.com"),
            Identifier(u"nothing", u"bob", u"example.com"),
            Value(u'custom.message.type',
                     "Some message contents"))
        self.runQueue(self.aliceQueue)
        [err] = self.flushLoggedErrors(MessageTransportError)


    def test_messageRemoteItem(self):
        """
        When a message is sent to a shared item using
        L{AMPMessenger.messageRemote}, and received by an L{AMPReceiver}, its
        data should be delivered to an appropriately decorated method.

        This is an integration test of the functionality covered by this suite
        and the functionality covered by L{AMPMessagingTests}.
        """
        aliceAMP = RealAMPReceiver(store=self.aliceStore)
        getEveryoneRole(self.aliceStore).shareItem(aliceAMP, u'ally')
        bobAMP = RealAMPReceiver(store=self.bobStore)
        getEveryoneRole(self.bobStore).shareItem(bobAMP, u'bobby')
        msgr = AMPMessenger(self.aliceQueue,
                            Identifier(u'ally', u'alice', u'example.com'),
                            Identifier(u'bobby', u'bob', u'example.com'))
        msgr.messageRemote(SimpleCommand, int1=3, str2="hello")
        self.runQueue(self.aliceQueue)
        self.assertEqual(bobAMP.args, [(3, 'hello')])



class SimpleError(Exception):
    """
    A simple error that should be detectable by L{SimpleCommand}.
    """



class SimpleCommand(Command):
    """
    Sample simple command with a few arguments.
    """

    arguments = [("int1", Integer()),
                 ("str2", String())]

    response = [("int3", Integer())]

    errors = {SimpleError: "SIMPLE_ERROR"}



class SenderCommand(Command):
    arguments = [("sender", SenderArgument())]
    response = [("sender", SenderArgument())]



class TargetCommand(Command):
    arguments = [("target", TargetArgument())]
    response = [("target", TargetArgument())]



class TrivialCommand(Command):
    """
    Trivial no-argument AMP command.
    """

    errors = {SimpleError: "SIMPLE_ERROR"}



class RealAMPReceiver(Item, AMPReceiver):
    """
    This is an integration testing item for making sure that decorated methods
    on items receiving messages will work.
    """

    dummy = integer()
    args = inmemory()

    def activate(self):
        """
        Set up test state.
        """
        self.args = []

    @commandMethod.expose(SimpleCommand)
    def doit(self, int1, str2):
        """
        Simple responder for L{SimpleCommand}.
        """
        self.args.append((int1, str2))
        return dict(int3=int1+len(str2))



class MyAMPReceiver(AMPReceiver):
    """
    A simple L{AMPReceiver} subclass with a few exposed methods.
    """

    def __init__(self):
        """
        Create a L{MyAMPReceiver}.
        """
        self.commandArguments = []
        self.commandAnswers = []
        self.commandErrors = []
        self.senders = []
        self.targets = []


    @commandMethod.expose(SimpleCommand)
    def simpleCommand(self, int1, str2):
        """
        Responder for a simple command.
        """
        self.commandArguments.append((int1, str2))
        return dict(int3=4)


    @answerMethod.expose(SimpleCommand)
    def simpleAnswer(self, int3):
        """
        Responder for a simple answer.
        """
        self.commandAnswers.append(int3)


    @errorMethod.expose(SimpleCommand, SimpleError)
    def simpleError(self, failure):
        self.commandErrors.append(failure)


    @commandMethod.expose(SenderCommand)
    def commandWithSender(self, sender):
        self.senders.append(sender)
        return {}


    @commandMethod.expose(TargetCommand)
    def commandWithTarget(self, target):
        self.targets.append(target)
        return {}

    @answerMethod.expose(SenderCommand)
    def answerWithSender(self, sender):
        self.senders.append(sender)

    @answerMethod.expose(TargetCommand)
    def answerWithTarget(self, target):
        self.targets.append(target)


class ExpectedBuggyReceiver(AMPReceiver):
    """
    This AMP responder will raise an expected exception type.
    """

    @commandMethod.expose(TrivialCommand)
    def raiseSimpleError(self):
        raise SimpleError("simple description")



class UnexpectedBuggyReceiver(AMPReceiver):
    """
    This AMP responder will raise an unexpected exception type.
    """

    @commandMethod.expose(TrivialCommand)
    def raiseRuntimeError(self):
        raise RuntimeError()



class AMPMessagingTests(TestCase):
    """
    Test cases for high-level AMP message parsing and emitting API.
    """

    def setUp(self):
        """
        Initialize the list of messages which this can deliver as a
        pseudo-queue.
        """
        self.messages = []


    def queueMessage(self, sender, target, value,
                     consequence=None):
        """
        Emulate L{MessageQueue.queueMessage}.
        """
        self.messages.append((sender, target,
                              value.type, value.data, consequence))


    def test_messageRemote(self):
        """
        L{AMPMessenger.messageRemote} should queue a message with the provided
        queue, sender, and target, serializing its arguments according to the
        provided AMP command.
        """
        sender = Identifier(u'test-sender', u'bob', u'example.com')
        target = Identifier(u'test-target', u'alice', u'example.com')

        msgr = AMPMessenger(self, sender, target)
        msgr.messageRemote(SimpleCommand, int1=3, str2="hello")
        expectedConsequence = None
        self.assertEqual(self.messages,
                         [(sender, target, AMP_MESSAGE_TYPE,
                           Box(_command="SimpleCommand",
                               int1="3", str2="hello").serialize(),
                           expectedConsequence)])


    def test_messageReceived(self):
        """
        L{AMPReceiver.messageReceived} should dispatch commands to methods that
        were appropriately decorated.
        """
        amr = MyAMPReceiver()
        questionBox = Box(_command=SimpleCommand.commandName,
                          int1="7", str2="test")
        data = questionBox.serialize()
        response = amr.messageReceived(
            Value(AMP_MESSAGE_TYPE, data), None, None)
        self.assertEqual(response.type, AMP_ANSWER_TYPE)
        self.assertEqual(amr.commandArguments, [(7, "test")])
        self.assertEqual(response.data, Box(int3="4").serialize())


    def test_messageReceivedHandledError(self):
        """
        L{AMPReceiver.messageReceived} should emit a responseType and
        responseData of the appropriate type if the command in question has a
        translation of the raised error.
        """
        bug = ExpectedBuggyReceiver()
        questionBox = Box(_command=TrivialCommand.commandName)
        data = questionBox.serialize()
        rar = self.assertRaises(RevertAndRespond, bug.messageReceived,
                                Value(AMP_MESSAGE_TYPE, data),
                                None, None)
        self.assertEqual(rar.value.type, AMP_ANSWER_TYPE)
        self.assertEqual(rar.value.data,
                         Box(_error_code="SIMPLE_ERROR",
                             _error_description="simple description")
                         .serialize())


    def test_messageReceivedUnhandledError(self):
        """
        L{AMPReceiver.messageReceived} should allow error not defined by its
        command to be raised so that the normal L{ERROR_REMOTE_EXCEPTION}
        behavior takes over.
        """
        bug = UnexpectedBuggyReceiver()
        questionBox = Box(_command=TrivialCommand.commandName)
        data = questionBox.serialize()
        self.assertRaises(RuntimeError, bug.messageReceived,
                          Value(AMP_MESSAGE_TYPE, data),
                          None, None)


    def test_answerReceived(self):
        """
        L{AMPReceiver.answerReceived} should dispatch answers to methods that
        were appropriately decorated.
        """
        originalMessageData = Box(_command=SimpleCommand.commandName).serialize()
        amr = MyAMPReceiver()
        answerBox = Box(int3="4")
        data = answerBox.serialize()
        amr.answerReceived(Value(AMP_ANSWER_TYPE, data),
                           Value(None, originalMessageData),
                           None, None)
        self.assertEqual(amr.commandAnswers, [4])


    def test_errorReceived(self):
        """
        L{AMPReceiver.answerReceived} should dispatch answers that indicate an
        AMP error to methods decorated by the L{errorMethod} decorator, not to
        the L{answerMethod} decorator.
        """
        originalMessageData = Box(_command=SimpleCommand.commandName).serialize()
        amr = MyAMPReceiver()
        data = Box(_error="SIMPLE_ERROR").serialize()
        amr.answerReceived(Value(AMP_ANSWER_TYPE, data),
                           Value(None, originalMessageData),
                           None, None)
        self.assertEqual(amr.commandAnswers, [])
        amr.commandErrors.pop().trap(SimpleError)
        self.assertEqual(amr.commandErrors, [])


    def test_messageReceivedWrongType(self):
        """
        An L{UnknownMessageType} should be raised when a message of the wrong
        type is dispatched to an L{AMPReceiver}.
        """
        amr = MyAMPReceiver()
        questionBox = Box(_command=SimpleCommand.commandName,
                          int1="7", str2="test")
        data = questionBox.serialize()
        for badType in u'some.random.type', AMP_ANSWER_TYPE:
            self.assertRaises(UnknownMessageType, amr.messageReceived,
                              Value(badType, data), None, None)
        self.assertEqual(amr.commandArguments, [])


    def test_messageReceivedBadData(self):
        """
        A L{MalformedMessage} should be raised when a message that cannot be
        interpreted as a single AMP box is received.
        """

        amr = MyAMPReceiver()
        for badData in ["", Box().serialize() + Box().serialize()]:
            self.assertRaises(MalformedMessage, amr.messageReceived,
                              Value(AMP_MESSAGE_TYPE, badData), None, None)


    def test_answerReceivedBadData(self):
        """
        A L{MalformedMessage} should be raised when a message that cannot be
        interpreted as a single AMP box is received.
        """
        originalMessageData = Box(_command=SimpleCommand.commandName).serialize()
        amr = MyAMPReceiver()
        for badData in ["", Box().serialize() + Box().serialize()]:
            self.assertRaises(MalformedMessage, amr.answerReceived,
                              Value(AMP_ANSWER_TYPE, badData),
                              Value(None, originalMessageData),
                              None, None)


    def test_answerReceivedWrongType(self):
        """
        An L{UnknownMessageType} exception should be raised when a answer of
        the wrong type is dispatched to an L{AMPReceiver}.
        """
        originalMessageData = Box(_command=SimpleCommand.commandName).serialize()
        amr = MyAMPReceiver()
        answerBox = Box(int3="4")
        data = answerBox.serialize()
        for badType in u'some.random.type', AMP_MESSAGE_TYPE:
            self.assertRaises(UnknownMessageType, amr.answerReceived,
                              Value(badType, data),
                              Value(None, originalMessageData),
                              None, None)
        self.assertEqual(amr.commandAnswers, [])


    def test_messageReceivedSenderArgument(self):
        """
        The special argument L{TargetArgument} should cause the L{sender}
        argument to L{messageReceived} to be passed as an argument to the
        responder method.
        """
        amr = MyAMPReceiver()
        shareident = Identifier(u'abc', u'def', u'ghi')
        amr.messageReceived(
            Value(AMP_MESSAGE_TYPE,
                     Box(_command=SenderCommand.commandName).serialize()),
            shareident, None)
        self.assertEqual([shareident], amr.senders)


    def test_messageReceivedTargetArgument(self):
        """
        The special argument L{TargetArgument} should cause the L{sender}
        argument to L{messageReceived} to be passed as an argument to the
        responder method.
        """
        amr = MyAMPReceiver()
        shareident = Identifier(u'abc', u'def', u'ghi')
        amr.messageReceived(Value(AMP_MESSAGE_TYPE,
                                     Box(_command=TargetCommand.commandName).serialize()),
                            None, shareident)
        self.assertEqual([shareident], amr.targets)


    def test_answerReceivedSenderArgument(self):
        """
        The special argument L{SenderArgument} should cause the
        L{originalSender} argument to L{answerReceived} to be passed as an
        argument to the responder method.
        """
        amr = MyAMPReceiver()
        shareident = Identifier(u'abc', u'def', u'ghi')
        amr.answerReceived(
            Value(AMP_ANSWER_TYPE, Box().serialize()),
            Value(None,
                     Box(_command=SenderCommand.commandName).serialize()),
            shareident, None)
        self.assertEqual([shareident], amr.senders)


    def test_answerReceivedTargetArgument(self):
        """
        The special argument L{TargetArgument} should cause the
        L{originalTarget} argument to L{answerReceived} to be passed as an
        argument to the responder method.
        """
        amr = MyAMPReceiver()
        shareident = Identifier(u'abc', u'def', u'ghi')
        amr.answerReceived(
            Value(AMP_ANSWER_TYPE, Box().serialize()),
            Value(None,
                     Box(_command=TargetCommand.commandName).serialize()),
            None,
            shareident)
        self.assertEqual([shareident], amr.targets)



class ExpositionTests(TestCase):
    """
    Tests for exposing methods with the L{_AMPExposer.expose} decorator, and
    retrieving them with the L{_AMPExposer.responderForName} lookup method.
    """

    def setUp(self):
        """
        Set up a local L{_AMPExposer} instance for testing.
        """
        self.ampExposer = _AMPExposer("amp exposer for testing")
        self.errorExposer = _AMPErrorExposer("lulz")


    def test_exposeCommand(self):
        """
        Exposing a method as a command object ought to make it accessible to
        L{responderForName}, and add a matching C{command} attribute to that
        result.
        """
        class TestClass(object):
            def __init__(self, x):
                self.num = x

            @self.ampExposer.expose(TrivialCommand)
            def thunk(self):
                return 'hi', self.num + 1

        tc = TestClass(3)
        callable = self.ampExposer.responderForName(tc, TrivialCommand.commandName)
        self.assertEqual(callable(), ("hi", 4))
        self.assertIdentical(callable.command, TrivialCommand)


    def test_exposeError(self):
        """
        A method exposed as an error handler for a particular type of error
        should be able to be looked up by the combination of the command and
        the error.
        """
        class TestClass(object):
            @self.errorExposer.expose(SimpleCommand, SimpleError)
            def thunk(self):
                raise SimpleError()
        tc = TestClass()
        thunk = self.errorExposer.errbackForName(tc, SimpleCommand.commandName, "SIMPLE_ERROR")
        self.assertEqual(thunk.exception, SimpleError)
        self.assertEqual(thunk.command, SimpleCommand)
        self.assertRaises(SimpleError, thunk)


    def test_exposeOnTwoTypes(self):
        """
        An integration test with L{epsilon.expose}; sanity checking to make
        sure that exposing different methods on different classes for the same
        command name yields the same results.
        """
        class TestClass(object):
            @self.ampExposer.expose(TrivialCommand)
            def thunk(self):
                return 1
        class TestClass2:
            @self.ampExposer.expose(TrivialCommand)
            def funk(self):
                return 2

        tc2 = TestClass2()
        callable = self.ampExposer.responderForName(tc2, TrivialCommand.commandName)
        self.assertEqual(callable(), 2)
        tc = TestClass()
        callable = self.ampExposer.responderForName(tc, TrivialCommand.commandName)
        self.assertEqual(callable(), 1)



# -*- test-case-name: xmantissa.test.test_webnav -*-

from epsilon.structlike import record

from zope.interface import implements

from nevow.inevow import IQ
from nevow import url

from nevow.stan import NodeNotFound

from xmantissa.ixmantissa import ITab
from xmantissa.fragmentutils import dictFillSlots

class TabMisconfiguration(Exception):
    def __init__(self, info, tab):
        Exception.__init__(
            self,
            "Inconsistent tab item factory information",
            info, tab)

TabInfo = record('priority storeID children linkURL authoritative',
                 authoritative=None)

class Tab(object):
    """
    Represent part or all of the layout of a single navigation tab.

    @ivar name: This tab's name.

    @type storeID: C{int}
    @ivar storeID: The Axiom store identifier of the Item to which the user
    should be directed when this tab is activated.

    @ivar priority: A float between 0 and 1 indicating the relative ordering of
    this tab amongst its peers.  Higher priorities sort sooner.

    @ivar children: A tuple of tabs beneath this one.

    @ivar authoritative: A flag indicating whether this instance of the
    conceptual tab with this name takes precedent over any other instance of
    the conceptual tab with this name.  It is an error for two instances of the
    same conceptual tab to be authoritative.

    @type linkURL: C{NoneType} or C{str}
    @ivar linkURL: If not C{None}, the location to which the user should be
    directed when this tab is activated.  This will override whatever value
    is supplied for C{storeID}.
    """

    _item = None
    implements(ITab)

    def __init__(self, name, storeID, priority, children=(),
                 authoritative=True, linkURL=None):
        self.name = name
        self.storeID = storeID
        self.priority = priority
        self.children = tuple(children)
        self.authoritative = authoritative
        self.linkURL = linkURL

    def __repr__(self):
        return '<%s%s %r/%0.3f %r [%r]>' % (self.authoritative and '*' or '',
                                            self.__class__.__name__,
                                            self.name,
                                            self.priority,
                                            self.storeID,
                                            self.children)

    def __iter__(self):
        raise TypeError("%r are not iterable" % (self.__class__.__name__,))

    def __getitem__(self, key):
        """Retrieve a sub-tab from this tab by name.
        """
        tabs = [t for t in self.children if t.name == key]
        assert len(tabs) < 2, "children mis-specified for " + repr(self)
        if tabs:
            return tabs[0]
        raise KeyError(key)

    def pathFromItem(self, item, avatar):
        """
        @param item: A thing that we linked to, and such.

        @return: a list of [child, grandchild, great-grandchild, ...] that
        indicates a path from me to the navigation for that item, or [] if
        there is no path from here to there.
        """
        for subnav in self.children:
            subpath = subnav.pathFromItem(item, avatar)
            if subpath:
                subpath.insert(0, self)
                return subpath
        else:
            myItem = self.loadForAvatar(avatar)
            if myItem is item:
                return [self]
        return []

def getTabs(navElements):
    # XXX TODO: multiple levels of nesting, this is hard-coded to 2.
    # Map primary tab names to a TabInfo
    primary = {}

    # Merge tab information from all nav plugins into one big structure
    for plg in navElements:
        for tab in plg.getTabs():
            if tab.name not in primary:
                primary[tab.name] = TabInfo(
                    priority=tab.priority,
                    storeID=tab.storeID,
                    children=list(tab.children),
                    linkURL=tab.linkURL)
            else:
                info = primary[tab.name]

                if info.authoritative:
                    if tab.authoritative:
                        raise TabMisconfiguration(info, tab)
                else:
                    if tab.authoritative:
                        info.authoritative = True
                        info.priority = tab.priority
                        info.storeID = tab.storeID
                        info.linkURL = tab.linkURL
                info.children.extend(tab.children)

    # Sort the tabs and their children by their priority
    def key(o):
        return -o.priority

    resultTabs = []

    for (name, info) in primary.iteritems():
        info.children.sort(key=key)

        resultTabs.append(
            Tab(name, info.storeID, info.priority, info.children,
                linkURL=info.linkURL))

    resultTabs.sort(key=key)

    return resultTabs

def setTabURLs(tabs, webTranslator):
    """
    Sets the C{linkURL} attribute on each L{Tab} instance
    in C{tabs} that does not already have it set

    @param tabs: sequence of L{Tab} instances
    @param webTranslator: L{xmantissa.ixmantissa.IWebTranslator}
                          implementor

    @return: None
    """

    for tab in tabs:
        if not tab.linkURL:
            tab.linkURL = webTranslator.linkTo(tab.storeID)
        setTabURLs(tab.children, webTranslator)

def getSelectedTab(tabs, forURL):
    """
    Returns the tab that should be selected when the current
    resource lives at C{forURL}.  Call me after L{setTabURLs}

    @param tabs: sequence of L{Tab} instances
    @param forURL: L{nevow.url.URL}

    @return: L{Tab} instance
    """

    flatTabs = []

    def flatten(tabs):
        for t in tabs:
            flatTabs.append(t)
            flatten(t.children)

    flatten(tabs)
    forURL = '/' + forURL.path

    for t in flatTabs:
        if forURL == t.linkURL:
            return t

    flatTabs.sort(key=lambda t: len(t.linkURL), reverse=True)

    for t in flatTabs:
        if not t.linkURL.endswith('/'):
            linkURL = t.linkURL + '/'
        else:
            linkURL = t.linkURL

        if forURL.startswith(linkURL):
            return t



def startMenu(translator, navigation, tag):
    """
    Drop-down menu-style navigation view.

    For each primary navigation element available, a copy of the I{tab}
    pattern will be loaded from the tag.  It will have its I{href} slot
    filled with the URL for that navigation item.  It will have its I{name}
    slot filled with the user-visible name of the navigation element.  It
    will have its I{kids} slot filled with a list of secondary navigation
    for that element.

    For each secondary navigation element available beneath each primary
    navigation element, a copy of the I{subtabs} pattern will be loaded
    from the tag.  It will have its I{kids} slot filled with a self-similar
    structure.

    @type translator: L{IWebTranslator} provider
    @type navigation: L{list} of L{Tab}

    @rtype: {nevow.stan.Tag}
    """
    setTabURLs(navigation, translator)
    getp = IQ(tag).onePattern

    def fillSlots(tabs):
        for tab in tabs:
            if tab.children:
                kids = getp('subtabs').fillSlots('kids', fillSlots(tab.children))
            else:
                kids = ''

            yield dictFillSlots(getp('tab'), dict(href=tab.linkURL,
                                                  name=tab.name,
                                                  kids=kids))
    return tag.fillSlots('tabs', fillSlots(navigation))



def settingsLink(translator, settings, tag):
    """
    Render the URL of the settings page.
    """
    return tag[translator.linkTo(settings.storeID)]



# This is somewhat redundant with startMenu.  The selected/not feature of this
# renderer should be added to startMenu and then templates can just use that
# and this can be deleted.
def applicationNavigation(ctx, translator, navigation):
    """
    Horizontal, primary-only navigation view.

    For the navigation element currently being viewed, copies of the
    I{selected-app-tab} and I{selected-tab-contents} patterns will be
    loaded from the tag.  For all other navigation elements, copies of the
    I{app-tab} and I{tab-contents} patterns will be loaded.

    For either case, the former pattern will have its I{name} slot filled
    with the name of the navigation element and its I{tab-contents} slot
    filled with the latter pattern.  The latter pattern will have its
    I{href} slot filled with a link to the corresponding navigation
    element.

    The I{tabs} slot on the tag will be filled with all the
    I{selected-app-tab} or I{app-tab} pattern copies.

    @type ctx: L{nevow.context.WebContext}
    @type translator: L{IWebTranslator} provider
    @type navigation: L{list} of L{Tab}

    @rtype: {nevow.stan.Tag}
    """
    setTabURLs(navigation, translator)
    selectedTab = getSelectedTab(navigation,
                                 url.URL.fromContext(ctx))

    getp = IQ(ctx.tag).onePattern
    tabs = []

    for tab in navigation:
        if tab == selectedTab or selectedTab in tab.children:
            p = 'selected-app-tab'
            contentp = 'selected-tab-contents'
        else:
            p = 'app-tab'
            contentp = 'tab-contents'

        childTabs = []
        for subtab in tab.children:
            try:
                subtabp = getp("subtab")
            except NodeNotFound:
                continue
            childTabs.append(
                dictFillSlots(subtabp, {
                        'name': subtab.name,
                        'href': subtab.linkURL,
                        'tab-contents': getp("subtab-contents")
                        }))
        tabs.append(dictFillSlots(
                getp(p),
                {'name': tab.name,
                 'tab-contents': getp(contentp).fillSlots(
                        'href', tab.linkURL),
                 'subtabs': childTabs}))

    ctx.tag.fillSlots('tabs', tabs)
    return ctx.tag



from axiom.store import Store
from axiom.dependency import installOn

from nevow.livetrial.testcase import TestCase
from nevow.athena import expose

from xmantissa import prefs


class _PrefMixin(object):
    def getWidgetDocument(self):
        s = Store()

        self.dpc = prefs.DefaultPreferenceCollection(store=s)
        installOn(s, self.dpc)

        f = prefs.PreferenceCollectionFragment(self.dpc)
        class Tab:
            name = ''
            children = ()

        f.tab = Tab
        f.setFragmentParent(self)
        return f


class GeneralPrefs(_PrefMixin, TestCase):
    """
    Test case which renders L{xmantissa.ixmantissa.DefaultPreferenceCollection}
    and ensures that values changed client-side are correctly persisted
    """
    jsClass = u'Mantissa.Test.GeneralPrefs'

    def checkPersisted(self, itemsPerPage, timezone):
        """
        Assert that our preference collection has had its C{itemsPerPage}
        and C{timezone} attributes set to C{itemsPerPage} and C{timezone}.
        Called after the deferred returned by the liveform controller's
        C{submit} method has fired
        """
        self.assertEquals(self.dpc.itemsPerPage, itemsPerPage)
        self.assertEquals(self.dpc.timezone, timezone)
    expose(checkPersisted)


class PrefCollection(_PrefMixin, TestCase):
    """
    Test case which renders L{xmantissa.ixmantissa.DefaultPreferenceCollection}
    and makes sure that the form appears after submit.
    """
    jsClass = u'Mantissa.Test.PrefCollectionTestCase'



import random

maxLongInt = (2**63)-1

def genkey():
    """
    Generate a key (an int representing 16 bytes of random data).
    """
    return random.randint(0, maxLongInt)

def _swapat(key, n):
    # n: 0-8; a swap is an 8-bit value; 2 ints between 0-15
    return (key >> ((8 * (n))+4)) & 0xf, (key >> (8 * n)) & 0xf

def _swap(l, a, b):
    l[a], l[b] = l[b], l[a]

def webIDToStoreID(key, webid):
    """
    Takes a webid (a 16-character str suitable for including in URLs) and a key
    (an int, a private key for decoding it) and produces a storeID.
    """
    if len(webid) != 16:
        return None
    try:
        int(webid, 16)
    except TypeError:
        return None
    except ValueError:
        return None
    l = list(webid)
    for nybbleid in range(7, -1, -1):
        a, b = _swapat(key, nybbleid)
        _swap(l, b, a)
    i = int(''.join(l), 16)
    return i ^ key

def storeIDToWebID(key, storeid):
    """
    Takes a key (int) and storeid (int) and produces a webid (a 16-character
    str suitable for including in URLs)
    """
    i = key ^ storeid
    l = list('%0.16x' % (i,))
    for nybbleid in range(0, 8):
        a, b = _swapat(key, nybbleid)
        _swap(l, a, b)
    return ''.join(l)

def _test():
    for y in range(100):
        key = genkey()
        print 'starting with key', key, hex(key)
        for sid in range(1000):
            wid = storeIDToWebID(key, sid)
            sid2 = webIDToStoreID(key, wid)
            assert sid == sid2, '%s != %s [%r %r]' % (sid, sid2, wid, key)
            #print wid, '<=>', sid, ':', hex(key)


def _simpletest():
    key = 0xfedcba9876543210
    web = storeIDToWebID(key, 100)
    print 'web', repr(web)
    sid = webIDToStoreID(key, web)
    print 'store', hex(sid)

if __name__ == '__main__':
    _test()




"""
Test the upgrade of L{UserInfo} from schema version 1 to 2, which collapsed the
I{firstName} attribute and I{lastName} attribute into a single I{realName}
attribute.
"""

from axiom.test.historic.stubloader import StubbedTest

from xmantissa.signup import UserInfo
from xmantissa.test.historic.stub_userinfo1to2 import FIRST, LAST

class UserInfoTests(StubbedTest):
    def test_realName(self):
        """
        L{UserInfo.realName} on the upgraded item should be set to the old
        values for the I{firstName} and I{lastName} attribute, separated by a
        space.
        """
        infoItem = self.store.findUnique(UserInfo)
        self.assertEqual(infoItem.realName, FIRST + u" " + LAST)



from axiom.test.historic import stubloader
from xmantissa.settings import Settings

class SettingsTestCase(stubloader.StubbedTest):
    def testUpgrade(self):
        self.assertEquals(self.store.count(Settings), 0)




"""
Tests for the upgrade of L{PrivateApplication} schema from 3 to 4.
"""

from axiom.userbase import LoginSystem
from axiom.test.historic.stubloader import StubbedTest

from xmantissa.ixmantissa import ITemplateNameResolver, IWebViewer
from xmantissa.website import WebSite
from xmantissa.webapp import PrivateApplication
from xmantissa.publicweb import CustomizedPublicPage
from xmantissa.webgestalt import AuthenticationApplication
from xmantissa.prefs import PreferenceAggregator, DefaultPreferenceCollection
from xmantissa.search import SearchAggregator

from xmantissa.test.historic.stub_privateApplication3to4 import (
    USERNAME, DOMAIN, PREFERRED_THEME, PRIVATE_KEY)

class PrivateApplicationUpgradeTests(StubbedTest):
    """
    Tests for L{xmantissa.webapp.privateApplication3to4}.
    """
    def setUp(self):
        d = StubbedTest.setUp(self)
        def siteStoreUpgraded(ignored):
            loginSystem = self.store.findUnique(LoginSystem)
            account = loginSystem.accountByAddress(USERNAME, DOMAIN)
            self.subStore = account.avatars.open()
            return self.subStore.whenFullyUpgraded()
        d.addCallback(siteStoreUpgraded)
        return d


    def test_powerup(self):
        """
        At version 4, L{PrivateApplication} should be an
        L{ITemplateNameResolver} powerup on its store.
        """
        application = self.subStore.findUnique(PrivateApplication)
        powerups = list(self.subStore.powerupsFor(ITemplateNameResolver))
        self.assertIn(application, powerups)


    def test_webViewer(self):
        """
        At version 5, L{PrivateApplication} should be an
        L{IWebViewer} powerup on its store.
        """
        application = self.subStore.findUnique(PrivateApplication)
        interfaces = list(self.subStore.interfacesFor(application))
        self.assertIn(IWebViewer, interfaces)


    def test_attributes(self):
        """
        All of the attributes of L{PrivateApplication} should have the same
        values on the upgraded item as they did before the upgrade.
        """
        application = self.subStore.findUnique(PrivateApplication)
        self.assertEqual(application.preferredTheme, PREFERRED_THEME)
        self.assertEqual(application.privateKey, PRIVATE_KEY)

        website = self.subStore.findUnique(WebSite)
        self.assertIdentical(application.website, website)

        customizedPublicPage = self.subStore.findUnique(CustomizedPublicPage)
        self.assertIdentical(
            application.customizedPublicPage, customizedPublicPage)

        authenticationApplication = self.subStore.findUnique(
            AuthenticationApplication)
        self.assertIdentical(
            application.authenticationApplication, authenticationApplication)

        preferenceAggregator = self.subStore.findUnique(PreferenceAggregator)
        self.assertIdentical(
            application.preferenceAggregator, preferenceAggregator)

        defaultPreferenceCollection = self.subStore.findUnique(
            DefaultPreferenceCollection)
        self.assertIdentical(
            application.defaultPreferenceCollection,
            defaultPreferenceCollection)

        searchAggregator = self.subStore.findUnique(SearchAggregator)
        self.assertIdentical(application.searchAggregator, searchAggregator)

        self.assertIdentical(application.privateIndexPage, None)
